"""Symlinking functionality"""

import fnmatch
import itertools
import json
import logging
import os
from collections import defaultdict
from collections.abc import Iterable, Sequence
from pathlib import Path

from . import filesystem as fs
from .inventory import load_ender_chest, load_ender_chest_instances, load_shulker_boxes
from .loggers import IMPORTANT, INVENTORY_LOGGER, PLACE_LOGGER
from .prompt import prompt
from .shulker_box import ShulkerBox


def place_ender_chest(
    minecraft_root: Path,
    keep_broken_links: bool = False,
    keep_stale_links: bool = False,
    error_handling: str = "abort",
    relative: bool = True,
    rollback: bool = False,
) -> dict[str, dict[Path, list[str]]]:
    """Link all instance files and folders to all shulker boxes

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    keep_broken_links : bool, optional
        By default, this method will remove any broken links in your instances
        and servers folders. To disable this behavior, pass in
        `keep_broken_links=True`.
    keep_stale_links : bool, optional
        By default, this method will remove any links into your EnderChest folder
        that are no longer specified by any shulker box (such as because the
        instance spec or shulker box configuration changed). To disable this
        behavior, pass in `keep_stale_links=True`.
    error_handling : str, optional
        By default, if a linking failure occurs, this method will terminate
        immediately (`error_handling=abort`). Alternatively,
          - pass in `error_handling="ignore"` to continue as if the link failure
            hadn't occurred
          - pass in `error_handling="skip"` to abort linking the current instance
            to the current shulker box but otherwise continue on
          - pass in `error_handling="skip-instance"` to abort linking the current
            instance altogether but to otherwise continue on with other instances
          - pass in `error_handling="skip-shulker-box"` to abort linking to the current
            shulker box altogether but to otherwise continue on with other boxes
          - pass in `error_handling="prompt"` to ask what to do on each failure
    relative : bool, optional
        By default, links will use relative paths when possible. To use absolute
        paths instead (see: https://bugs.mojang.com/projects/MC/issues/MC-263046),
        pass in `relative=False`. See note below.
    rollback: bool, optional
        In the future in the event of linking errors passing in `rollback=True`
        can be used to roll back any changes that have already been applied
        based on the error-handling method specified.

    Returns
    -------
    dict
        A record of the placed symlinks, structured as a nested dict:

        - the top-level keys are the instance names, with the values being a map
          of the links placed within those instances:
            - the keys of those mappings are the relative paths of the placed
              symlinks inside the instance folder
            - the values are the list of shulker boxes, sorted in ascending
              priority, into which that symlink was linked (explicitly, the
              _last_ entry in each list corresponds to the shulker box inside
              which that link currently points)

    Notes
    -----
    - If one of the files or folders being placed is itself a symlink, relative
      links will be created as *nested* links (a link pointing to the link),
      whereas in "absolute" mode (`relative=False`), the link that will be
      placed will point **directly** to the final target
    - This can lead to the stale-link cleanup behavior not correctly removing
      an outdated symlink if the fully resolved target of a link falls outside
      the EnderChest folder
    - The generated placement record reflects only the placements performed by
      _this_ placement operation ("stale" links will never be included)
    - The generated placements record will include broken links irrespective of
      the `keep_broken_links` argument
    - If the placement is aborted (`error_handling="abort"` or "Abort" selected
      from prompt) then the returned placements will be empty
    """
    placements: dict[str, dict[Path, list[str]]] = {}

    if rollback is not False:  # pragma: no cover
        raise NotImplementedError("Rollbacks are not currently supported")

    try:
        host = load_ender_chest(minecraft_root).name
    except (FileNotFoundError, ValueError) as bad_chest:
        PLACE_LOGGER.error(
            "Could not load EnderChest from %s:\n  %s", minecraft_root, bad_chest
        )
        return {}

    instances = load_ender_chest_instances(minecraft_root, log_level=logging.DEBUG)

    shulker_boxes: list[ShulkerBox] = []

    for shulker_box in load_shulker_boxes(minecraft_root, log_level=logging.DEBUG):
        if not shulker_box.matches_host(host):
            PLACE_LOGGER.debug(
                "%s is not intended for linking to this host (%s)",
                shulker_box.name,
                host,
            )
            continue
        shulker_boxes.append(shulker_box)

    skip_boxes: list[ShulkerBox] = []

    def handle_error(shulker_box: ShulkerBox | None) -> str:
        """Centralized error-handling

        Parameters
        ----------
        shulker_box:
            The current shulker box (in case it needs to be added to the skip list)

        Returns
        -------
        str
            Instructions on what to do next. Options are:
              - retry
              - return
              - break
              - continue
              - pass
        """
        if error_handling == "prompt":
            proceed_how = (
                prompt(
                    "How would you like to proceed?"
                    "\n[Q]uit; [R]etry; [C]ontinue; skip linking the rest of this:"
                    "\n[I]nstance, [S]hulker box, shulker/instance [M]atch?",
                    suggestion="R",
                )
                .lower()
                .replace(" ", "")
                .replace("-", "")
                .replace("_", "")
            )
            match proceed_how:
                case "" | "r":
                    proceed_how = "retry"
                case "" | "i" | "instance" | "skipinstance":
                    proceed_how = "skip-instance"
                case "q" | "quit" | "abort" | "exit" | "stop":
                    proceed_how = "abort"
                case "c" | "continue" | "ignore":
                    proceed_how = "ignore"
                case "m" | "match" | "skip":
                    proceed_how = "skip"
                case "s" | "shulker" | "shulkerbox" | "skipshulker":
                    proceed_how = "skip-shulker"
                case _:
                    PLACE_LOGGER.error("Invalid selection.")
                    return handle_error(shulker_box)
        else:
            proceed_how = error_handling

        match proceed_how:
            case "retry":
                return "retry"
            case "abort" | "stop" | "quit" | "exit":
                PLACE_LOGGER.error("Aborting")
                return "return"
            case "ignore":
                PLACE_LOGGER.debug("Ignoring")
                return "pass"
            case "skip":
                PLACE_LOGGER.warning("Skipping the rest of this match")
                return "continue"
            case "skip-instance":
                PLACE_LOGGER.warning("Skipping any more linking from this instance")

                return "break"
            case "skip-shulker-box" | "skip-shulkerbox" | "skip-shulker":
                PLACE_LOGGER.warning("Skipping any more linking into this shulker box")
                if shulker_box:
                    skip_boxes.append(shulker_box)
                return "continue"
            case _:
                raise ValueError(
                    f"Unrecognized error-handling method: {error_handling}"
                )

    for instance in instances:
        instance_root = (minecraft_root / instance.root.expanduser()).expanduser()
        placements[instance.name] = defaultdict(list)

        handling: str | None = "retry"
        while handling == "retry":
            if instance_root.exists():
                handling = None
                break

            PLACE_LOGGER.error(
                "No minecraft instance exists at %s",
                instance_root.expanduser().absolute(),
            )
            handling = handle_error(None)
        if handling is not None:
            match handling:
                case "return":
                    return {}  # intentionally wipe the cache
                case "break":
                    break
                case _:  # nothing to link, so might as well skip the rest
                    continue

        # start by removing all existing symlinks into the EnderChest
        if not keep_stale_links:
            for file in instance_root.rglob("*"):
                if file.is_symlink():
                    if fs.links_into_enderchest(minecraft_root, file):
                        PLACE_LOGGER.debug(
                            "Removing old link: %s -> %s", file, os.readlink(file)
                        )
                        file.unlink()

        for shulker_box in shulker_boxes:
            if not shulker_box.matches(instance):
                continue
            if shulker_box in skip_boxes:
                continue

            box_root = shulker_box.root.expanduser().absolute()

            PLACE_LOGGER.info("Linking %s to %s", instance.root, shulker_box.name)

            resources = set(_rglob(box_root, shulker_box.max_link_depth))

            match_exit = "pass"
            for link_folder in shulker_box.link_folders:
                resources -= {box_root / link_folder}
                resources -= set((box_root / link_folder).rglob("*"))

                handling = "retry"
                while handling == "retry":
                    try:
                        link_resource(link_folder, box_root, instance_root, relative)
                        placements[instance.name][Path(link_folder)].append(
                            shulker_box.name
                        )
                        handling = None
                    except OSError:
                        PLACE_LOGGER.error(
                            "Error linking shulker box %s to instance %s:"
                            "\n  %s is a non-empty directory",
                            shulker_box.name,
                            instance.name,
                            (instance.root / link_folder),
                        )
                        handling = handle_error(shulker_box)
                if handling is not None:
                    match handling:
                        case "return":
                            return placements
                        case "break":
                            match_exit = "break"
                            break
                        case "continue":
                            match_exit = "continue"
                            break
                        case "pass":
                            continue  # or pass--it's the end of the loop

            if match_exit not in ("break", "continue"):
                for resource in resources:
                    resource_path = resource.relative_to(box_root)
                    for pattern in shulker_box.do_not_link:
                        if fnmatch.fnmatchcase(
                            str(resource_path), pattern
                        ) or fnmatch.fnmatchcase(
                            str(resource_path), os.path.join("*", pattern)
                        ):
                            PLACE_LOGGER.debug(
                                "Skipping %s (matches pattern %s)",
                                resource_path,
                                pattern,
                            )
                            break
                    else:
                        handling = "retry"
                        while handling == "retry":
                            try:
                                link_resource(
                                    resource_path,
                                    box_root,
                                    instance_root,
                                    relative,
                                )
                                placements[instance.name][resource_path].append(
                                    shulker_box.name
                                )
                                handling = None
                            except OSError:
                                PLACE_LOGGER.error(
                                    "Error linking shulker box %s to instance %s:"
                                    "\n  %s already exists",
                                    shulker_box.name,
                                    instance.name,
                                    instance.root / resource_path,
                                )
                                handling = handle_error(shulker_box)
                        if handling is not None:
                            match handling:
                                case "return":
                                    return placements
                                case "break":
                                    match_exit = "break"
                                    break
                                case "continue":
                                    match_exit = "continue"  # technically does nothing
                                    break
                                case "pass":
                                    continue  # or pass--it's the end of the loop

            # consider this a "finally"
            if not keep_broken_links:
                # we clean up as we go, just in case of a failure
                for file in instance_root.rglob("*"):
                    if not file.exists():
                        PLACE_LOGGER.debug("Removing broken link: %s", file)
                        file.unlink()

            if match_exit == "break":
                break
    return placements


def link_resource(
    resource_path: str | Path,
    shulker_root: Path,
    instance_root: Path,
    relative: bool,
) -> None:
    """Create a symlink for the specified resource from an instance's space
    pointing to the tagged file / folder living inside a shulker box.

    Parameters
    ----------
    resource_path : str or Path
        Location of the resource relative to the instance's ".minecraft" folder
    shulker_root : Path
        The path to the shulker box
    instance_root : Path
        The path to the instance's ".minecraft" folder
    relative : bool
        If True, the link will be use a relative path if possible. Otherwise,
        an absolute path will be used, regardless of whether a relative or
        absolute path was provided.

    Raises
    ------
    OSError
        If a file or non-empty directory already exists where you're attempting
        to place the symlink

    Notes
    -----
    - This method will create any folders that do not exist within an instance
    - This method will overwrite existing symlinks and empty folders
      but will not overwrite or delete any actual files.
    """
    instance_path = (instance_root / resource_path).expanduser().absolute()
    instance_path.parent.mkdir(parents=True, exist_ok=True)

    target: str | Path = (shulker_root / resource_path).expanduser().absolute()
    if relative:
        target = os.path.relpath(target, instance_path.parent)
    else:
        target = target.resolve()  # type: ignore

    if instance_path.is_symlink():
        # remove previous symlink in this spot
        instance_path.unlink()
        PLACE_LOGGER.debug("Removed previous link at %s", instance_path)
    else:
        try:
            os.rmdir(instance_path)
            PLACE_LOGGER.debug("Removed empty directory at %s", instance_path)
        except FileNotFoundError:
            pass  # A-OK

    PLACE_LOGGER.debug("Linking %s to %s", instance_path, target)
    os.symlink(
        target,
        instance_path,
        target_is_directory=(shulker_root / resource_path).is_dir(),
    )


def _rglob(root: Path, max_depth: int) -> Iterable[Path]:
    """Find all files (and directories* and symlinks) in the path up to the
    specified depth

    Parameters
    ----------
    root : Path
        The path to search
    max_depth : int
        The maximum number of levels to go

    Returns
    -------
    list-like of paths
        The files (and directories and symlinks) in the path up to that depth

    Notes
    -----
    - Unlike an actual rglob, this method does not return any directories that
      are not at the maximum depth
    - Setting max_depth to 0 (or below) will return all files in the root, but
      ***be warned*** that because this method follows symlinks, you can very
      easily find yourself in an infinite loop
    """
    top_level = root.iterdir()
    if max_depth == 1:
        return top_level
    return itertools.chain(
        *(
            _rglob(path, max_depth - 1) if path.is_dir() else (path,)
            for path in top_level
        )
    )


def cache_placements(
    minecraft_root: Path, placements: dict[str, dict[Path, list[str]]]
) -> None:
    """Write placement record to file

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    placements : dict
        A record of placed links, as generated by `place_ender_chest`
    """
    cache_file = fs.place_cache(minecraft_root)
    cache_file.write_text(
        json.dumps(
            {
                instance_name: {
                    str(resource_path): shulker_boxes
                    for resource_path, shulker_boxes in instance_placements.items()
                }
                for instance_name, instance_placements in placements.items()
            },
            indent=4,
            sort_keys=False,
        )
    )
    PLACE_LOGGER.debug("Placement cache written to %s", cache_file)


def load_placement_cache(minecraft_root: Path) -> dict[str, dict[Path, list[str]]]:
    """Load the placement cache from file

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)

    Returns
    -------
    dict
        A record of the placed symlinks, structured as a nested dict, matching
        the schema of one generated by `place_ender_chest`

        Raises
    ------
    OSError
        If the placement cache could not be found, read or parsed
    """
    try:
        cache_file = fs.place_cache(minecraft_root)
        INVENTORY_LOGGER.debug(
            "Loading placement cache from %s", fs.place_cache(minecraft_root)
        )
        raw_dict: dict[str, dict[str, list[str]]] = json.loads(
            cache_file.read_text("UTF-8")
        )
    except json.JSONDecodeError as decode_error:
        raise OSError(
            f"{fs.place_cache(minecraft_root)} is corrupted and could not be parsed:"
        ) from decode_error
    return {
        instance_name: {
            Path(resource_path): shulker_boxes
            for resource_path, shulker_boxes in instance_placements.items()
        }
        for instance_name, instance_placements in raw_dict.items()
    }


def trace_resource(
    minecraft_root: Path,
    pattern: str,
    placements: dict[str, dict[Path, list[str]]],
    instance_name: str | None = None,
) -> list[tuple[Path, Path, list[str]]]:
    """Given a filename or glob pattern, return a list of all matching
    EnderChest-placed symlinks, together with a trace-back of the shulker boxes
    each link targets

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    pattern : filename, path or glob pattern
        The resource to trace
    placements : dict
        A record of placed symlinks, such as the one generated by `place_ender_chest`.
    instance_name : str, optional
        The name of the instance to search. This variable is case-sensitive.
        If None is given, all instances will be searched.

    Returns
    -------
    list of (Path, Path, list) tuples
        - The first item in each list is the path of the instance root
        - The second item in each list is the path to a linked resource
          matching the provided pattern (and instance), relative to the instance
          root
        - The third item is the list of shulker boxes, sorted in ascending
          priority, into which that symlink was linked (explicitly, the
          _last_ entry in each list corresponds to the shulker box inside
          which that link currently points)

    Raises
    ------
    OSError
        If no placement cache was provided and the placement cache file could
        not be found, read or parsed
    KeyError
        If there is no instance registered to this EnderChest with the specified
        name
    """
    instances = {
        instance.name: instance
        for instance in load_ender_chest_instances(
            minecraft_root, log_level=logging.DEBUG
        )
    }
    if instance_name is None:
        return sum(
            (
                trace_resource(minecraft_root, pattern, placements, name)
                for name in instances
            ),
            [],
        )
    instance_root = instances[instance_name].root
    matches: list[tuple[Path, Path, list[str]]] = []
    for resource_path, target_boxes in placements[instance_name].items():
        if (
            fnmatch.fnmatchcase(str(resource_path), pattern)
            or fnmatch.fnmatchcase(str(resource_path), os.path.join("*", pattern))
            or fnmatch.fnmatchcase(
                os.path.abspath(minecraft_root / instance_root / resource_path),
                os.path.join("*", pattern),
            )
        ):
            matches.append((instance_root, resource_path, target_boxes))
    return matches


def report_resource_trace(
    minecraft_root: Path, instance_root: Path, resource_path: Path, boxes: Sequence[str]
) -> None:
    """Print (log) the shulker boxes an instance resource is linked to

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    instance_root : Path
        The path of the EnderChest-placed symlink
    resource_path : Path
        The path to the symlink, relative to the instance root
    boxes : list of str
        The names of the shulker boxes, sorted by ascending priority, that are
        targeted by this symlink (technically only the last entry in this list
        is the actual target)
    """
    symlink_location = instance_root / resource_path
    if len(boxes) == 0:  # pragma: no cover
        # Since defaultdicts are involved, this could happen accidentally at
        # some point and should just be ignored
        return
    *other_box_names, primary_box_name = boxes
    try:
        INVENTORY_LOGGER.log(
            IMPORTANT,
            "%s currently resolves to %s",
            symlink_location,
            os.path.abspath(
                (
                    symlink_location / (minecraft_root / symlink_location).readlink()
                ).expanduser()
            ),
        )
    except OSError:
        INVENTORY_LOGGER.warning(
            "%s no longer exists or is not a symlink", symlink_location
        )

    INVENTORY_LOGGER.log(
        IMPORTANT,
        "    based on being linked into shulker box: %s",
        primary_box_name,
    )
    INVENTORY_LOGGER.debug(
        "        - > %s",
        fs.shulker_box_root(minecraft_root, primary_box_name) / resource_path,
    )

    for box_name in reversed(other_box_names):
        INVENTORY_LOGGER.info(
            "    which overwrote the link into shulker box: %s", box_name
        )
        INVENTORY_LOGGER.debug(
            "        - > %s",
            fs.shulker_box_root(minecraft_root, box_name) / resource_path,
        )


def list_placements(
    minecraft_root: Path, pattern: str, instance_name: str | None = None
) -> None:
    """Report all shulker boxes that provide files matching the given pattern

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    pattern : filename, path or glob pattern
        The pattern of the resource to trace
    instance_name : str, optional
        The name of the instance to search. This variable is case-sensitive.
        If None is given, all instances will be searched.
    """
    try:
        placements = load_placement_cache(minecraft_root)
    except OSError as no_cache:
        INVENTORY_LOGGER.error(
            "The placement cache could not be loaded:"
            "\n  %s"
            "\nPlease run enderchest place again to regenerate the cache.",
            no_cache,
        )
        return
    try:
        matches = trace_resource(
            minecraft_root, pattern, placements, instance_name=instance_name
        )
    except KeyError:
        INVENTORY_LOGGER.error(
            "No instance named %s is registered to this EnderChest", instance_name
        )
        return
    if len(matches) == 0:
        INVENTORY_LOGGER.warning(
            "Could not find any placed resources matching the pattern %s%s."
            "\n\nNote: this command does not check inside linked folders.",
            pattern,
            f"\nin the instance {instance_name}" if instance_name else "",
        )
        return
    for match in matches:
        report_resource_trace(minecraft_root, *match)



"""Functionality for resolving EnderChest and shulker box states"""

import logging
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any
from urllib.parse import ParseResult

from enderchest.sync import render_remote

from . import EnderChest, InstanceSpec, ShulkerBox
from . import filesystem as fs
from .loggers import INVENTORY_LOGGER


def load_ender_chest(minecraft_root: Path) -> EnderChest:
    """Load the configuration from the enderchest.cfg file in the EnderChest
    folder.

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)

    Returns
    -------
    EnderChest
        The EnderChest configuration

    Raises
    ------
    FileNotFoundError
        If no EnderChest folder exists in the given minecraft root or if no
        enderchest.cfg file exists within that EnderChest folder
    ValueError
        If the EnderChest configuration is invalid and could not be parsed
    """
    config_path = fs.ender_chest_config(minecraft_root)
    INVENTORY_LOGGER.debug("Loading %s", config_path)
    ender_chest = EnderChest.from_cfg(config_path)
    INVENTORY_LOGGER.debug("Parsed EnderChest installation from %s", minecraft_root)
    return ender_chest


def load_ender_chest_instances(
    minecraft_root: Path, log_level: int = logging.INFO
) -> Sequence[InstanceSpec]:
    """Get the list of instances registered with the EnderChest located in the
    minecraft root

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    log_level : int, optional
        By default, this method will report out the minecraft instances it
        finds at the INFO level. You can optionally pass in a lower (or higher)
        level if this method is being called from another method where that
        information is redundant or overly verbose.

    Returns
    -------
    list of InstanceSpec
        The instances registered with the EnderChest

    Notes
    -----
    If no EnderChest is installed in the given location, then this will return
    an empty list rather than failing outright.
    """
    try:
        ender_chest = load_ender_chest(minecraft_root)
        instances: Sequence[InstanceSpec] = ender_chest.instances
    except (FileNotFoundError, ValueError) as bad_chest:
        INVENTORY_LOGGER.error(
            "Could not load EnderChest from %s:\n  %s", minecraft_root, bad_chest
        )
        instances = []
    if len(instances) == 0:
        INVENTORY_LOGGER.warning(
            "There are no instances registered to the %s EnderChest", minecraft_root
        )
    else:
        INVENTORY_LOGGER.log(
            log_level,
            "These are the instances that are currently registered"
            " to the %s EnderChest:\n%s",
            minecraft_root,
            "\n".join(
                [
                    f"  {i + 1}. {render_instance(instance)}"
                    for i, instance in enumerate(instances)
                ]
            ),
        )
    return instances


def render_instance(instance: InstanceSpec) -> str:
    """Render an instance spec to a descriptive string

    Parameters
    ----------
    instance : InstanceSpec
        The instance spec to render

    Returns
    -------
    str
        {instance.name} ({instance.root})
    """
    return f"{instance.name} ({instance.root})"


def load_shulker_boxes(
    minecraft_root: Path, log_level: int = logging.INFO
) -> list[ShulkerBox]:
    """Load all shulker boxes in the EnderChest folder and return them in the
    order in which they should be linked.

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    log_level : int, optional
        By default, this method will report out the minecraft instances it
        finds at the INFO level. You can optionally pass in a lower (or higher)
        level if this method is being called from another method where that
        information is redundant or overly verbose.

    Returns
    -------
    list of ShulkerBoxes
        The shulker boxes found in the EnderChest folder, ordered in terms of
        the sequence in which they should be linked

    Notes
    -----
    If no EnderChest is installed in the given location, then this will return
    an empty list rather than failing outright.
    """
    shulker_boxes: list[ShulkerBox] = []

    try:
        for shulker_config in fs.shulker_box_configs(minecraft_root):
            try:
                shulker_boxes.append(_load_shulker_box(shulker_config))
            except (FileNotFoundError, ValueError) as bad_shulker:
                INVENTORY_LOGGER.warning(
                    "%s\n  Skipping shulker box %s",
                    bad_shulker,
                    shulker_config.parent.name,
                )

    except FileNotFoundError:
        INVENTORY_LOGGER.error(
            "There is no EnderChest installed within %s",
            minecraft_root,
        )
        return []

    shulker_boxes = sorted(shulker_boxes)

    if len(shulker_boxes) == 0:
        if log_level >= logging.INFO:
            INVENTORY_LOGGER.warning(
                "There are no shulker boxes within the %s EnderChest", minecraft_root
            )
    else:
        report_shulker_boxes(
            shulker_boxes, log_level, f"the {minecraft_root} EnderChest"
        )
    return shulker_boxes


def report_shulker_boxes(
    shulker_boxes: Iterable[ShulkerBox], log_level: int, ender_chest_name: str
) -> None:
    """Log the list of shulker boxes in the order they'll be linked

    Parameters
    ----------
    shulker_boxes : list of ShulkerBoxes
        The shulker boxes to report on
    log_level : int
        The log level of the report
    ender_chest_name : str
        Which chest is this?

    Returns
    -------
    None
    """
    # TODO: properly log minecraft_root when one appears in the message
    INVENTORY_LOGGER.log(
        log_level,
        f"These are the shulker boxes within {ender_chest_name}"
        "\nlisted in the order in which they are linked:\n%s",
        "\n".join(
            f"  {shulker_box.priority}. {_render_shulker_box(shulker_box)}"
            for shulker_box in shulker_boxes
        ),
    )


def _load_shulker_box(config_file: Path) -> ShulkerBox:
    """Attempt to load a shulker box from a config file, and if you can't,
    at least log why the loading failed.

    Parameters
    ----------
    config_file : Path
        Path to the config file

    Returns
    -------
    ShulkerBox | None
        The parsed shulker box or None, if the shulker box couldn't be parsed

    Raises
    ------
    FileNotFoundError
        If the given config file could not be found
    ValueError
        If there was a problem parsing the config file
    """
    INVENTORY_LOGGER.debug("Attempting to parse %s", config_file)
    shulker_box = ShulkerBox.from_cfg(config_file)
    INVENTORY_LOGGER.debug("Successfully parsed %s", _render_shulker_box(shulker_box))
    return shulker_box


def _render_shulker_box(shulker_box: ShulkerBox) -> str:
    """Render a shulker box to a descriptive string

    Parameters
    ----------
    shulker_box : ShulkerBox
        The shulker box spec to render

    Returns
    -------
    str
        {priority}. {folder_name} [({name})]
            (if different from folder name)
    """
    stringified = f"{shulker_box.root.name}"
    if shulker_box.root.name != shulker_box.name:  # pragma: no cover
        # note: this is not a thing
        stringified += f" ({shulker_box.name})"
    return stringified


def load_ender_chest_remotes(
    minecraft_root: Path, log_level: int = logging.INFO
) -> list[tuple[ParseResult, str]]:
    """Load all remote EnderChest installations registered with this one

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    log_level : int, optional
        By default, this method will report out the minecraft instances it
        finds at the INFO level. You can optionally pass in a lower (or higher)
        level if this method is being called from another method where that
        information is redundant or overly verbose.

    Returns
    -------
    list of (URI, str) tuples
        The URIs of the remote EnderChests, paired with their aliases

    Notes
    -----
    If no EnderChest is installed in the given location, then this will return
    an empty list rather than failing outright.
    """
    try:
        ender_chest = load_ender_chest(minecraft_root)
        remotes: Sequence[tuple[ParseResult, str]] = ender_chest.remotes
    except (FileNotFoundError, ValueError) as bad_chest:
        INVENTORY_LOGGER.error(
            "Could not load EnderChest from %s:\n  %s", minecraft_root, bad_chest
        )
        remotes = ()

    if len(remotes) == 0:
        if log_level >= logging.INFO:
            INVENTORY_LOGGER.warning(
                "There are no remotes registered to the %s EnderChest", minecraft_root
            )
        return []

    report = (
        "These are the remote EnderChest installations registered"
        " to the one installed at %s"
    )
    remote_list: list[tuple[ParseResult, str]] = []
    log_args: list[Any] = [minecraft_root]
    for remote, alias in remotes:
        report += "\n  - %s"
        log_args.append(render_remote(alias, remote))
        remote_list.append((remote, alias))
    INVENTORY_LOGGER.log(log_level, report, *log_args)
    return remote_list


def get_shulker_boxes_matching_instance(
    minecraft_root: Path, instance_name: str
) -> list[ShulkerBox]:
    """Get the list of shulker boxes that the specified instance links to

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    instance_name : str
        The name of the instance you're asking about

    Returns
    -------
    list of ShulkerBox
        The shulker boxes that are linked to by the specified instance
    """
    try:
        chest = load_ender_chest(minecraft_root)
    except (FileNotFoundError, ValueError) as bad_chest:
        INVENTORY_LOGGER.error(
            "Could not load EnderChest from %s:\n  %s", minecraft_root, bad_chest
        )
        return []
    for mc in chest.instances:
        if mc.name == instance_name:
            break
    else:
        INVENTORY_LOGGER.error(
            "No instance named %s is registered to this EnderChest", instance_name
        )
        return []

    matches = [
        box
        for box in load_shulker_boxes(minecraft_root, log_level=logging.DEBUG)
        if box.matches(mc) and box.matches_host(chest.name)
    ]

    if len(matches) == 0:
        report = "does not link to any shulker boxes in this chest"
    else:
        # TODO: render as args
        report = "links to the following shulker boxes:\n" + "\n".join(
            f"  - {_render_shulker_box(box)}" for box in matches
        )

    INVENTORY_LOGGER.info("The instance %s %s", render_instance(mc), report)

    return matches


def get_instances_matching_shulker_box(
    minecraft_root: Path, shulker_box_name: str
) -> list[InstanceSpec]:
    """Get the list of registered instances that link to the specified shulker box

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    shulker_box_name : str
        The name of the shulker box you're asking about

    Returns
    -------
    list of InstanceSpec
        The instances that are / should be linked to the specified shulker box
    """
    try:
        config_file = fs.shulker_box_config(minecraft_root, shulker_box_name)
    except FileNotFoundError:
        INVENTORY_LOGGER.error("No EnderChest is installed in %s", minecraft_root)
        return []
    try:
        shulker_box = _load_shulker_box(config_file)
    except (FileNotFoundError, ValueError) as bad_box:
        INVENTORY_LOGGER.error(
            "Could not load shulker box %s\n  %s", shulker_box_name, bad_box
        )
        return []

    chest = load_ender_chest(minecraft_root)

    if not shulker_box.matches_host(chest.name):
        INVENTORY_LOGGER.warning(
            "This shulker box will not link to any instances on this machine"
        )
        return []

    if not chest.instances:
        INVENTORY_LOGGER.warning(
            "This EnderChest does not have any instances registered."
            " To register some, run the command:"
            "\nenderchest gather minecraft",
        )
        return []

    INVENTORY_LOGGER.debug(
        "These are the instances that are currently registered"
        " to the %s EnderChest:\n%s",
        minecraft_root,
        "\n".join(
            [
                f"  {i + 1}. {render_instance(instance)}"
                for i, instance in enumerate(chest.instances)
            ]
        ),
    )

    matches = [
        instance for instance in chest.instances if shulker_box.matches(instance)
    ]

    if len(matches) == 0:
        report = "is not linked to by any registered instances"
    else:
        report = "is linked to by the following instances:\n" + "\n".join(
            f"  - {render_instance(instance)}" for instance in matches
        )

    INVENTORY_LOGGER.info(
        "The shulker box %s %s", _render_shulker_box(shulker_box), report
    )

    return matches



"""Tests around file discovery and registration"""

import os
from pathlib import Path

import pytest

from enderchest import craft
from enderchest import filesystem as fs
from enderchest import gather
from enderchest import instance as i
from enderchest import inventory

from . import utils


class TestGatherInstances:
    @pytest.fixture(autouse=True)
    def setup_teardown(self, minecraft_root, home):
        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False)
        )

        yield

        # tests that with offer-to-update set to False, EnderChest doesn't go rogue
        assert not (home / ".minecraft" / "allowed_symlinks.txt").exists()

    def test_official_instance_parsing(self, home):
        assert utils.normalize_instance(
            gather.gather_metadata_for_official_instance(home / ".minecraft")
        ) == utils.normalize_instance(utils.TESTING_INSTANCES[0])

    def test_instance_search_finds_official_instance(self, minecraft_root, home):
        assert i.equals(
            minecraft_root,
            gather.gather_minecraft_instances(minecraft_root, home, official=True)[0],
            utils.TESTING_INSTANCES[0],
        )

    @pytest.mark.parametrize(
        "instance, idx, dot_root",
        (
            ("axolotl", 1, True),
            ("bee", 2, True),
            ("chest-boat", 3, False),
        ),
    )
    def test_mmc_instance_parsing(self, minecraft_root, instance, idx, dot_root):
        assert utils.normalize_instance(
            gather.gather_metadata_for_mmc_instance(
                minecraft_root
                / "instances"
                / instance
                / (".minecraft" if dot_root else "minecraft")
            )
        ) == utils.normalize_instance(
            # we're not testing aliasing right now
            utils.TESTING_INSTANCES[idx]._replace(name=instance)
        )

    def test_instance_search_finds_mmc_instances(self, minecraft_root):
        instances = sorted(
            gather.gather_minecraft_instances(
                minecraft_root, minecraft_root, official=False
            ),
            key=lambda instance: instance.name,
        )

        assert len(instances) == 4

        assert all(
            [
                i.equals(
                    minecraft_root, instances[idx - 1], utils.TESTING_INSTANCES[idx]
                )
                for idx in range(1, 5)
            ]
        )

    def test_instance_search_doesnt_choke_on_relative_root(
        self, minecraft_root, monkeypatch
    ):
        monkeypatch.chdir(minecraft_root.parent)
        instances = sorted(
            gather.gather_minecraft_instances(
                Path(minecraft_root.name), Path(minecraft_root.name), official=False
            ),
            key=lambda instance: instance.name,
        )

        assert len(instances) == 4

        assert all(
            [
                i.equals(
                    minecraft_root, instances[idx - 1], utils.TESTING_INSTANCES[idx]
                )
                for idx in range(1, 5)
            ]
        )

    def test_instance_search_can_find_all_instances(self, minecraft_root, home):
        instances = sorted(
            gather.gather_minecraft_instances(
                minecraft_root, minecraft_root, official=None
            ),
            key=lambda instance: (
                instance.name if instance.name != "official" else "aaa"
            ),  # sorting hack
        )

        assert len(instances) == 5

        assert all(
            [
                i.equals(minecraft_root, instances[idx], utils.TESTING_INSTANCES[idx])
                for idx in range(5)
            ]
        )

    def test_instance_search_warns_if_no_instances_can_be_found(
        self, minecraft_root, caplog
    ):
        empty_folder = minecraft_root / "nothing in there"
        empty_folder.mkdir()
        instances = gather.gather_minecraft_instances(
            minecraft_root, empty_folder, official=None
        )

        assert len(instances) == 0

        warn_log = "\n".join(
            record.msg for record in caplog.records if record.levelname == "WARNING"
        )

        assert "Could not find any Minecraft instances" in warn_log

    def test_onboarding_new_instances(self, minecraft_root, home):
        # start with a blank chest
        craft.craft_ender_chest(minecraft_root, remotes=(), overwrite=True)
        ec_config = (
            fs.ender_chest_config(minecraft_root).read_text("utf-8").splitlines()
        )
        # this is a horrible way tp do this
        ec_config[6] = "offer-to-update-symlink-allowlist = False"
        fs.ender_chest_config(minecraft_root).write_text("\n".join(ec_config))
        assert inventory.load_ender_chest(minecraft_root).instances == ()
        gather.update_ender_chest(minecraft_root, (home, minecraft_root / "instances"))

        instances = sorted(
            inventory.load_ender_chest_instances(minecraft_root),
            key=lambda instance: (
                instance.name if instance.name != "official" else "aaa"
            ),  # sorting hack
        )

        assert len(instances) == 5

        assert all(
            [
                i.equals(minecraft_root, instances[idx], utils.TESTING_INSTANCES[idx])
                for idx in range(5)
            ]
        )

    def test_updating_an_instance_does_not_overwrite_tags(
        self, minecraft_root, home
    ) -> None:
        enderchest = inventory.load_ender_chest(minecraft_root)

        expected: dict[str, tuple[str, ...]] = {}
        for idx, instance in enumerate(enderchest._instances):
            tags = tuple((str(num) for num in range(idx + 1)))
            expected[instance.name] = tuple(sorted(tags + instance.groups_))
            enderchest._instances[idx] = instance._replace(
                tags_=tags, groups_=("outdated",)
            )
        enderchest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        gather.update_ender_chest(minecraft_root, (home, minecraft_root / "instances"))

        enderchest = inventory.load_ender_chest(minecraft_root)

        assert expected == {
            instance.name: instance.tags for instance in enderchest.instances
        }


class TestGatherServer:
    @pytest.fixture()
    def server_jars(self, minecraft_root: Path) -> dict[Path, i.InstanceSpec]:
        expected_metadata = {}
        server_folder = minecraft_root / "servers"
        server_folder.mkdir(exist_ok=True)

        # this might actually be Aether 1?
        (server_folder / "aether.2").mkdir()
        (
            server_folder
            / "aether.2"
            / "forge-1.7.10-10.13.4.1614-1.7.10-installer.jar"
        ).write_text("Wrong jar\n")
        (server_folder / "aether.2" / ".minecraft").mkdir()
        server_jar = (
            server_folder
            / "aether.2"
            / ".minecraft"
            / "forge-1.7.10-10.13.4.1614-1.7.10-universal.jar"
        )
        server_jar.write_text("Correct jar\n")
        expected_metadata[server_jar] = i.InstanceSpec(
            server_jar.parent.parent.name,
            server_jar.parent.parent,
            ("1.7.10",),
            "Forge",
            ("server",),
            (),
        )

        (
            server_folder / "aether.2" / ".minecraft" / "minecraft_server.1.7.10.jar"
        ).write_text("Wrong jar\n")
        (server_folder / "aether.2" / "mods").mkdir()
        (server_folder / "aether.2" / "mods" / "aether-1.7.10-1.6.jar").write_text(
            "Wrong jar\n"
        )

        (server_folder / "aether.legacy").mkdir()
        (
            server_folder / "aether.legacy" / "forge-1.12.2-14.23.5.2860-installer.jar"
        ).write_text("Wrong jar\n")
        (server_folder / "aether.legacy" / ".minecraft").mkdir()
        server_jar = (
            server_folder
            / "aether.legacy"
            / ".minecraft"
            / "forge-1.12.2-14.23.5.2860.jar"
        )
        server_jar.write_text("Correct jar\n")
        expected_metadata[server_jar] = i.InstanceSpec(
            server_jar.parent.parent.name,
            server_jar.parent.parent,
            ("1.12.2",),
            "Forge",
            ("server",),
            (),
        )
        (
            server_folder
            / "aether.legacy"
            / ".minecraft"
            / "minecraft_server.1.12.2.jar"
        ).write_text("Wrong jar\n")
        (server_folder / "aether.legacy" / "mods").mkdir()
        (
            server_folder / "aether.legacy" / "mods" / "aether-1.12.2-v1.5.2.jar"
        ).write_text("Wrong jar\n")

        (server_folder / "chunk.in.a.globe").mkdir()
        server_jar = (
            server_folder
            / "chunk.in.a.globe"
            / "fabric-server-mc.1.18.1-loader.0.14.4-launcher.0.10.2.jar"
        )
        server_jar.write_text("Correct jar\n")
        expected_metadata[server_jar] = i.InstanceSpec(
            server_jar.parent.name,
            server_jar.parent,
            ("1.18.1",),
            "Fabric Loader",
            ("server",),
            (),
        )
        (server_folder / "chunk.in.a.globe" / ".fabric" / "server").mkdir(parents=True)
        for jar_name in (
            "1.18.1-server.jar",
            "fabric-loader-server-0.13.3-minecraft-1.18.1.jar",
            "fabric-loader-server-0.14.4-minecraft-1.18.1.jar",
        ):
            (
                server_folder / "chunk.in.a.globe" / ".fabric" / "server" / jar_name
            ).write_text("Wrong jar\n")

        centralized_jar_folder = (
            minecraft_root / "EnderChest" / "Chest Monster" / "server_jars"
        )
        centralized_jar_folder.mkdir(parents=True, exist_ok=True)
        for jar_name, launcher in (
            ("minecraft_server-1.20.4.jar", ""),
            (
                "fabric-server-mc.1.20.4-loader.0.15.7-launcher.1.0.0.jar",
                "Fabric Loader",
            ),
            ("forge-1.20.4-49.0.30-shim.jar", "Forge"),
            ("paper-1.20.4-424.jar", "Paper"),
            ("purpur-1.20.4-2142.jar", "Purpur"),
            ("spigot-1.20.4.jar", "Spigot"),
        ):
            (centralized_jar_folder / jar_name).write_text("Correct jar\n")
            expected_metadata[centralized_jar_folder / jar_name] = i.InstanceSpec(
                "ignoreme", Path("ignoreme"), ("1.20.4",), launcher, ("server",), ()
            )
        return expected_metadata

    def test_parsing_metadata_from_jar(self, server_jars) -> None:
        expected: list[tuple[Path, tuple[str], str]] = [
            (jar, instance.minecraft_versions, instance.modloader)
            for jar, instance in server_jars.items()
        ]
        results: list[tuple[Path, tuple[str], str]] = []
        for jar in server_jars:
            meta = gather._gather_metadata_from_jar_filename(jar.name)
            results.append((jar, meta["minecraft_versions"], meta["modloader"]))

        assert expected == results

    @pytest.mark.parametrize(
        "server", ("aether.2", "aether.legacy", "chunk.in.a.globe")
    )
    def test_gather_server_instance_parses_metadata_from_the_correct_jar(
        self, server, server_jars, minecraft_root, monkeypatch, capsys
    ):
        script_reader = utils.scripted_prompt(["", ""])
        monkeypatch.setattr("builtins.input", script_reader)

        server_home = minecraft_root / "servers" / server
        instance_meta = gather.gather_metadata_for_minecraft_server(server_home)

        _ = capsys.readouterr()

        for jar, meta in server_jars.items():
            if jar.is_relative_to(server_home):
                expected_meta = meta
                break
        else:
            raise RuntimeError(
                "Whoops! Test class fixture doesn't contain meta for this server home."
            )

        assert expected_meta == instance_meta

    @pytest.mark.parametrize(
        "with_explicit_path", (False, True), ids=("", "with_explicit_path")
    )
    def test_mystery_jar_will_prompt_for_info(
        self, minecraft_root, with_explicit_path, monkeypatch, capsys
    ):
        script_reader = utils.scripted_prompt(["vanilla", "1.0", "OG", "old,vanilla"])
        monkeypatch.setattr("builtins.input", script_reader)
        server_jar = minecraft_root / "servers" / "mystery" / "server.jar"
        server_jar.parent.mkdir(parents=True)
        server_jar.write_text("Correct jar\n")

        instance_meta = gather.gather_metadata_for_minecraft_server(
            server_jar.parent, server_jar=server_jar if with_explicit_path else None
        )

        _ = capsys.readouterr()

        assert instance_meta == i.InstanceSpec(
            "OG",
            server_jar.parent,
            ("1.0",),
            "",
            ("server",),
            ("old", "vanilla"),
        )


class TestSymlinkAllowlistVersionChecker:
    """aka check my regex"""

    @pytest.mark.parametrize(
        "version_string, expected",
        (
            ("1.19.4", False),
            ("1.20", True),
            ("1.20.1-pre1", True),
            ("1.21-pre1", True),
            ("1.19-rc2", False),
            ("1.2.1", False),
            ("22w43a", False),
            ("23w9a", False),
            ("23w09a", False),
            ("23w51a", True),
            ("24w6a", True),
            ("24w06a", True),
        ),
    )
    def test_check_if_version_needs_allow_list(self, version_string, expected):
        assert gather._needs_symlink_allowlist(version_string) == expected


class TestSymlinkAllowlistHandling:
    @pytest.fixture(autouse=True)
    def setup_teardown(self, minecraft_root):
        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False)
        )
        with fs.ender_chest_config(minecraft_root).open("w") as ec_config:
            ec_config.write(
                """
[properties]
"""
            )

        utils.populate_mmc_instance_folder(
            minecraft_root / "instances" / "talespin" / "minecraft",
            "1.20",
            "Quilt",
            "talespin",
        )

        (
            minecraft_root
            / "instances"
            / "talespin"
            / "minecraft"
            / "allowed_symlinks.txt"
        ).write_text(
            "my_development_folder\n"
        )  # read: not EnderChest

        yield

        assert (
            (
                minecraft_root
                / "instances"
                / "talespin"
                / "minecraft"
                / "allowed_symlinks.txt"
            )
            .read_text()
            .startswith("my_development_folder\n")
        )

    @pytest.mark.parametrize("instance", ("official", "talespin"))
    def test_ender_chest_does_not_write_allowlist_without_consent(
        self, minecraft_root, home, monkeypatch, capsys, instance
    ):
        if instance == "official":
            search_path = home
        else:
            search_path = minecraft_root / "instances" / "talespin"
        i_forbid_it = utils.scripted_prompt(["no"])
        monkeypatch.setattr("builtins.input", i_forbid_it)

        gather.gather_minecraft_instances(minecraft_root, search_path, None)

        _ = capsys.readouterr()  # suppress outputs

        # easier to check both
        assert not (home / ".minecraft" / "allowed_symlinks.txt").exists()
        assert (
            minecraft_root
            / "instances"
            / "talespin"
            / "minecraft"
            / "allowed_symlinks.txt"
        ).read_text() == "my_development_folder\n"

        # make sure all responses were used
        with pytest.raises(StopIteration):
            i_forbid_it()

    @pytest.mark.parametrize(
        "symlinked_root", (False, True), ids=("direct_root", "symlinked_root")
    )
    def test_ender_chest_will_write_allowlists_with_consent(
        self, symlinked_root, minecraft_root, home, monkeypatch, capsys, tmp_path
    ):
        mkay = utils.scripted_prompt(["y"] * 3)
        monkeypatch.setattr("builtins.input", mkay)

        if symlinked_root:
            (tmp_path / "mc").symlink_to(minecraft_root)
            provided_root = tmp_path / "mc"
        else:
            provided_root = minecraft_root

        gather.gather_minecraft_instances(provided_root, home, True)

        # this is also testing that you're not getting prompted
        # for pre-1.20 instances
        gather.gather_minecraft_instances(
            provided_root, minecraft_root / "instances", False
        )

        _ = capsys.readouterr()  # suppress outputs

        ender_chest_path = os.path.realpath(fs.ender_chest_folder(minecraft_root))

        assert (
            home / ".minecraft" / "allowed_symlinks.txt"
        ).read_text() == ender_chest_path + "\n"
        assert (
            minecraft_root
            / "instances"
            / "talespin"
            / "minecraft"
            / "allowed_symlinks.txt"
        ).read_text() == f"my_development_folder\n{ender_chest_path}\n"

        # make sure all responses were used
        with pytest.raises(StopIteration):
            mkay()



"""rsync sync implementation. Relies on the user having rsync installed on their system"""

import os.path
import re
import shutil
import subprocess
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from urllib.parse import ParseResult, unquote

from . import SYNC_LOGGER, get_default_netloc, uri_to_ssh

RSYNC = shutil.which("rsync")
if RSYNC is None:  # pragma: no cover
    raise RuntimeError(
        "No rsync executable found on your system. Cannot sync using this protocol."
    )


def _get_rsync_version() -> tuple[int, int]:
    """Determine the installed version of Rsync

    Returns
    -------
    int
        The major version of the resolved Rsync executable
    int
        The minor version of the resolved Rsync executable

    Raises
    -----
    RuntimeError
        If Rsync is not installed, if `rsync --version` returns an error or if
        the version information cannot be decoded from the `rsync --version`
        output
    """
    try:
        result = subprocess.run(
            ["rsync", "--version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False,
        )
        if result.stderr:  # TODO: #124 just use check=True
            raise RuntimeError(result.stderr.decode("utf-8"))

        head = result.stdout.decode("utf-8").splitlines()[0]
    except (FileNotFoundError, IndexError):
        raise RuntimeError("Rsync is not installed or could not be executed.")

    try:
        if match := re.match(
            r"^rsync[\s]+version ([0-9]+).([0-9]+).([0-9]+)",
            head,
        ):
            major, minor, *_ = match.groups()
            return int(major), int(minor)
        raise AssertionError
    except (AssertionError, ValueError):
        raise RuntimeError(f"Could not parse version output:\n{head}")


rsync_version = _get_rsync_version()
if rsync_version < (3, 2):
    raise RuntimeError(
        "EnderChest requires Rsync 3.2 or newer."
        " The version detected on your system is {}.{}".format(*rsync_version)
    )


def run_rsync(
    working_directory: Path,
    source: str,
    destination_folder: str,
    delete: bool,
    dry_run: bool,
    exclude: Iterable[str],
    *additional_args: str,
    timeout: int | None = None,
    verbosity: int = 0,
    rsync_flags: str | None = None,
) -> None:
    """Run an operation with rsync

    Parameters
    ----------
    working_directory : Path
        The working directory to run the sync command from
    source : str
        The source file or folder to sync, specified as either a URI string,
        an ssh address or a path relative to the working directory
    destination_folder : str
        The destination folder where the file or folder should be synced to,
        with the same formats available as for source
    delete : bool
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    exclude : list of str
        Any patterns that should be excluded from the sync (and sync)
    *additional_args : str
        Any additional arguments to pass into the rsync command
    timeout : int, optional
        The number of seconds to wait before timing out the sync operation.
        If None is provided, no explicit timeout value will be set.
    verbosity : int
        A modifier for how much info to output either to stdout or the INFO-level
        logs. At...

          - verbosity = -2 : No information will be printed, even on dry runs
          - verbosity = -1 : The sync itself will be silent. Dry runs will only
                             report the sync statistics.
          - verbosity =  0 : Actual syncs will display a progress bar. Dry run
                             reports will summarize the changes to each shulker
                             box in addition to reporting the sync statistics .
          - verbosity =  1 : Actual syncs will report the progress of each file
                             transfer. Dry runs will report on each file to
                             be created, updated or deleted.
          - verbosity =  2 : Dry runs and syncs will print or log the output
                             of rsync run using the `-vv` modifier

        Verbosity values outside of this range will simply be capped / floored
        to [-2, 2].
    rsync_flags : str, optional
        By default, rsync will be run using the flags "shaz" which means:

          - no space splitting
          - use output (file sizes, mostly) human-readable
          - archive mode (see: https://www.baeldung.com/linux/rsync-archive-mode)
          - compress data during transfer

        Advanced users may choose to override these options, but **you do so
        at your own peril**.

    Raises
    ------
    TimeoutError
        If the rsync operation times out before completion
    RuntimeError
        If the rsync operation fails for any other reason

    Notes
    -----
    This method does not perform any validation or normalization of the source,
    destination, exclude-list, additional arguments or rsync options.
    """
    rsync_flags = rsync_flags or "shaz"

    args: list[str] = [RSYNC, f"-{rsync_flags}"]  # type: ignore[list-item]
    if delete:
        args.append("--delete")
    if dry_run:
        args.extend(("--dry-run", "--stats"))
        if verbosity < 1:
            # at 1+ we don't need it to be machine-parseable
            args.append("--out-format=%i %n")
    else:
        if verbosity >= 0:
            args.append("--stats")
        if verbosity == 0:
            args.append("--info=progress2")
        if verbosity >= 1:
            args.append("--progress")
    if verbosity > 0:
        args.append("-" + "v" * verbosity)

    for pattern in exclude:
        args.extend(("--exclude", pattern))
    args.extend(additional_args)
    args.extend((source, destination_folder))

    SYNC_LOGGER.debug(
        "Executing the following command:\n %s",
        " ".join(args),
    )

    with subprocess.Popen(
        args,
        stdout=subprocess.PIPE if dry_run else None,
        stderr=subprocess.PIPE,
        cwd=working_directory,
    ) as proc:
        if timeout:
            try:
                proc.wait(timeout)
            except subprocess.TimeoutExpired as times_up:
                proc.kill()
                if proc.stdout is not None:
                    if output_log := proc.stdout.read().decode("UTF-8"):
                        SYNC_LOGGER.warning(output_log)
                if proc.stderr is not None:
                    if error_log := proc.stderr.read().decode("UTF-8"):
                        SYNC_LOGGER.error(error_log)
                raise TimeoutError("Timeout reached.") from times_up

        if proc.stdout is not None:
            if output_log := proc.stdout.read().decode("UTF-8"):
                if verbosity > 0:
                    dry_run_output = output_log.splitlines()
                else:
                    dry_run_output = summarize_rsync_report(output_log)
                    SYNC_LOGGER.info("\nSUMMARY\n-------")
                for line in dry_run_output:
                    if _is_important_stats_line(line):
                        SYNC_LOGGER.log(25, line)
                    else:
                        SYNC_LOGGER.debug(line)

        if proc.stderr is not None:
            if error_log := proc.stderr.read().decode("UTF-8"):
                if "No such file or directory" in error_log:
                    raise FileNotFoundError(error_log)
                raise RuntimeError(error_log)  # pragma: no cover


def summarize_rsync_report(raw_output: str, depth: int = 2) -> list[str]:
    """Take the captured output from running
    `rsync -ha --out-format="%i %n"`
    and report a high-level summary to the logging.INFO level

    Parameters
    ----------
    raw_output : str
        The raw output captured from running the rsync command
    depth : int, optional
        How many directories to go down from the root to generate the summary.
        Default is 2 (just report on top-level files and folders within the
        source folder).

    Returns
    -------
    list of str
        Any lines that weren't part of the rsync report (and were probably
        part of `--stats`?)

    Notes
    -----
    The rsync man page (https://linux.die.net/man/1/rsync) describes the output
    format as... "cryptic," which I find rather charitable. The relevant bits
    are that `--out-format="%i %n"` produces:
    - `%i` : a string of 11 characters that gives various metadata about the file
      transfer operation (is it a file, a directory or a link? Is it being
      sent or received? Created, updated or deleted?)
    - `%n`: the path of the file (or whatever), unquoted, un-escaped
    """
    summary: dict[str, dict[str, int] | str] = defaultdict(
        lambda: {"create": 0, "update": 0, "delete": 0}
    )
    stats: list[str] = []
    for line in raw_output.splitlines():
        if line == "":  # skip empty lines
            continue

        info = line.split()[0]
        full_path = os.path.normpath(" ".join(line.split()[1:]))
        path_key = os.sep.join(full_path.split(os.sep)[:depth])

        if info.startswith("*deleting"):
            if full_path == path_key:
                summary[path_key] = "delete"
            else:
                entry = summary[path_key]
                if not isinstance(entry, str):
                    entry["delete"] += 1
                # otherwise the whole thing is being deleted
        elif info[2:5] == "+++":  # this is a creation
            if full_path == path_key:
                summary[path_key] = "create"
            else:
                if info[1] != "d":  # don't count directories
                    entry = summary[path_key]
                    if isinstance(entry, str):
                        # then this is described by the top-level op
                        pass
                    else:
                        entry["create"] += 1
                    # otherwise the whole key is being created
        elif info[:2] in ("<f", ">f"):  # file transfer
            # and remember that creates were caught above, so this must be an update
            if full_path == path_key:
                summary[path_key] = "update"
            else:
                entry = summary[path_key]
                if isinstance(entry, str):  # pragma: no cover
                    # this should never happen, but still
                    pass
                else:
                    entry["update"] += 1
        elif info[:2] == "cL":  # this is replacing a link, as far as I can tell
            if full_path == path_key:
                summary[path_key] = "update"
            else:
                entry = summary[path_key]
                if isinstance(entry, str):  # pragma: no cover
                    # this should never happen, but still
                    pass
                else:
                    entry["update"] += 1
        elif info[:1] == ".":  # pragma: no cover
            # this just means permissions or dates are being updated or something
            pass
        else:  # then hopefully this is part of the stats report
            stats.append(line)
            continue

        SYNC_LOGGER.debug(line)

    for path_key, report in sorted(summary.items()):
        if isinstance(report, str):
            # nice that these verbs follow the same pattern
            SYNC_LOGGER.info(f"{report[:-1].title()}ing %s", path_key)
        else:
            SYNC_LOGGER.info(
                "Within %s...\n%s",
                path_key,
                "\n".join(
                    f"  - {op[:-1].title()}ing {count} file{'' if count == 1 else 's'}"
                    for op, count in report.items()
                ),
            )
    return stats


def _is_important_stats_line(line: str) -> bool:
    """Determine if a stats line is worth logging at the INFO level (or whether
    it should be relegated to the DEBUG log)

    Parameters
    ----------
    line : str
        The log line to evaluate

    Returns
    -------
    bool
        True if and only if the line is identified as important
    """
    return line.startswith(
        (
            "Number of created files:",
            "Number of deleted files:",
            "Number of regular files transferred:",
            "Total transferred file size:",
        )
    )


def pull(
    remote_uri: ParseResult,
    local_path: Path,
    exclude: Iterable[str],
    dry_run: bool,
    use_daemon: bool = False,
    timeout: int | None = None,
    delete: bool = True,
    verbosity: int = 0,
    rsync_args: Iterable[str] | None = None,
) -> None:
    """Sync an upstream file or folder into the specified location using rsync.
    This will overwrite any files and folders already at the destination.

    Parameters
    ----------
    remote_uri : ParseResult
        The URI for the remote resource to copy from
    local_path : Path
        The destination folder
    exclude : list of str
        Any patterns that should be excluded from the sync
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    use_daemon : bool, optional
        By default, the rsync is performed over ssh. If you happen to have an
        rsync daemon running on your system, however, you're welcome to leverage
        it instead by passing in `use_daemon=True`
    timeout : int, optional
        The number of seconds to wait before timing out the sync operation.
        If None is provided, no explicit timeout value will be set.
    delete : bool, optional
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source. Default is True.
    verbosity : int
        A modifier for how much info to output either to stdout or the INFO-level
        logs. Defaults to 0.
    rsync_args: list of str, optional
        Any additional arguments to pass into rsync. Note that rsync is run by
        default with the flags: `-shaz`

    Raises
    ------
    FileNotFoundError
        If the destination folder does not exist

    Notes
    -----
    - This method does not provide for interactive authentication. If using
      rsync over SSH, you'll need to be set up for password-less (key-based)
      access.
    - If the destination folder does not already exist, this method will not
      create it or its parent directories.
    """
    if not local_path.exists():
        raise FileNotFoundError(f"{local_path} does not exist")

    if remote_uri.netloc == get_default_netloc():
        SYNC_LOGGER.debug("Performing sync as a local transfer")
        remote_path: str = unquote(remote_uri.path)
    elif use_daemon:
        remote_path = remote_uri.geturl()
    else:
        remote_path = uri_to_ssh(remote_uri)

    if rsync_args:  # pragma: no cover
        raise NotImplementedError

    run_rsync(
        local_path.parent,
        remote_path,
        local_path.name,
        delete,
        dry_run,
        exclude,
        *(rsync_args or ()),
        timeout=timeout,
        verbosity=verbosity,
    )


def push(
    local_path: Path,
    remote_uri: ParseResult,
    exclude: Iterable[str],
    dry_run: bool,
    use_daemon: bool = False,
    timeout: int | None = None,
    delete: bool = True,
    verbosity: int = 0,
    rsync_args: Iterable[str] | None = None,
) -> None:
    """Sync a local file or folder into the specified location using rsync.
    This will overwrite any files and folders already at the destination.

    Parameters
    ----------
    local_path : Path
        The file or folder to copy
    remote_uri : ParseResult
        The URI for the remote location to copy into
    exclude : list of str
        Any patterns that should be excluded from the sync
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    use_daemon : bool, optional
        By default, the rsync is performed over ssh. If you happen to have an
        rsync daemon running on your system, however, you're welcome to leverage
        it instead by passing in `use_daemon=True`
    timeout : int, optional
        The number of seconds to wait before timing out the sync operation.
        If None is provided, no explicit timeout value will be set.
    delete : bool, optional
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source. Default is True.
    verbosity : int
        A modifier for how much info to output either to stdout or the INFO-level
        logs. Defaults to 0.
    rsync_args: list of str, optional
        Any additional arguments to pass into rsync. Note that rsync is run by
        default with the flags: `-shaz`

    Notes
    -----
    - This method does not provide for interactive authentication. If using
      rsync over SSH, you'll need to be set up for password-less (key-based)
      access.
    - If the destination folder does not already exist, this method will very
      likely fail.
    """
    if remote_uri.netloc == get_default_netloc():
        SYNC_LOGGER.debug("Performing sync as a local transfer")
        remote_path: str = unquote(remote_uri.path)
    elif use_daemon:
        remote_path = remote_uri.geturl()
    else:
        remote_path = uri_to_ssh(remote_uri)

    if rsync_args:  # pragma: no cover
        raise NotImplementedError

    run_rsync(
        local_path.parent,
        local_path.name,
        remote_path,
        delete,
        dry_run,
        exclude,
        *(rsync_args or ()),
        timeout=timeout,
        verbosity=verbosity,
    )



"""Functionality for copying all files into their instances"""

import os
import shutil
from collections.abc import Iterable
from pathlib import Path

from . import filesystem as fs
from .enderchest import create_ender_chest
from .instance import InstanceSpec
from .inventory import load_ender_chest, load_ender_chest_instances
from .loggers import BREAK_LOGGER, IMPORTANT
from .prompt import confirm


def break_ender_chest(minecraft_root: Path) -> None:
    """Replace all instance symlinks with their actual targets, effectively
    "uninstalling" EnderChest

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    """
    instances = load_ender_chest_instances(minecraft_root, log_level=IMPORTANT)
    if not instances:
        BREAK_LOGGER.error("Aborting.")
        return

    BREAK_LOGGER.warning(
        "Are you sure you want to uninstall this EnderChest?"
        "\nDoing so will replace ALL the symlinks in each of the above instances"
        "\nwith copies of their EnderChest-linked targets."
        "\n\nTHIS CANNOT EASILY BE UNDONE!!"
    )
    if not confirm(default=False):
        BREAK_LOGGER.error("Aborting.")
        return

    chest_folder = fs.ender_chest_folder(minecraft_root)
    _break(chest_folder, instances)

    BREAK_LOGGER.log(
        IMPORTANT,
        "EnderChest has been uninstalled."
        "\nYou may now delete %s"
        "\nand uninstall the EnderChest package",
        chest_folder,
    )


def break_instances(minecraft_root: Path, instance_names: Iterable[str]) -> None:
    """Deregister the specified instances from EnderChest, replacing all
    instance symlinks with their actual targets, and then removing those
    instances from the enderchest.cfg

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    instance_names : list of str
        The names of the instances to break
    """
    ender_chest = load_ender_chest(minecraft_root)

    instance_lookup = {instance.name: instance for instance in ender_chest.instances}
    instances: list[InstanceSpec] = []
    for name in instance_names:
        try:
            instances.append(instance_lookup[name])
        except KeyError:
            BREAK_LOGGER.warning(
                'No instance named "%s" is registered to this EnderChest.'
                "\nSkipping.",
                name,
            )
    if len(instances) == 0:
        BREAK_LOGGER.error("No valid instances specified.\nAborting.")
        return

    BREAK_LOGGER.warning(
        "Are you sure you want to remove the following instances from your EnderChest?"
        "\n%s\nDoing so will replace ALL the symlinks in each of the above instances"
        "\nwith copies of their EnderChest-linked targets."
        "\n\nTHIS CANNOT EASILY BE UNDONE!!",
        "\n".join((f"  - {instance.name}" for instance in instances)),
    )
    if not confirm(default=False):
        BREAK_LOGGER.error("Aborting.")
        return

    _break(fs.ender_chest_folder(minecraft_root), instances)
    for instance in instances:
        ender_chest._instances.remove(instance)
    create_ender_chest(minecraft_root, ender_chest)


def _break(
    chest_folder: Path,
    instances: Iterable[InstanceSpec],
) -> None:
    """Actually perform the uninstallation

    Parameters
    ----------
    chest_folder : Path
        The path of the EnderChest folder that's being "broken"
    instances : list of InstanceSpec
        The instances to clear of EnderChest links
    """
    chest_folder = chest_folder.expanduser().resolve()

    for instance in instances:
        BREAK_LOGGER.info("Copying files into %s", instance.name)
        for resource_path in instance.root.expanduser().rglob("*"):
            if not resource_path.is_symlink():
                continue

            literal_target = resource_path.readlink()

            direct_target = Path(
                os.path.normpath(resource_path.readlink().expanduser().absolute())
            )

            final_target = resource_path.resolve().expanduser()

            if not direct_target.is_relative_to(
                chest_folder
            ) and not final_target.is_relative_to(chest_folder):
                # note: there's a pathological case where someone does something
                # silly like:
                # ~/.minecraft/options.txt -> ~/options.txt
                #    -> EnderChest/options.txt -> /configs/minecraft_options.txt
                # where deleting your EnderChest folder would break the chain,
                # but EnderChest wouldn't have ever created that chain, so that
                # person is on their own.
                continue

            try:
                resource_path.unlink()
                BREAK_LOGGER.debug(
                    "Removed link: %s -> %s", resource_path, literal_target
                )
                if final_target.is_relative_to(chest_folder):
                    if final_target.is_dir():
                        shutil.copytree(
                            final_target,
                            resource_path,
                            symlinks=True,
                            dirs_exist_ok=True,
                        )
                    else:
                        shutil.copy2(
                            final_target,
                            resource_path,
                            follow_symlinks=False,
                        )
                else:
                    resource_path.symlink_to(final_target)

                BREAK_LOGGER.debug("Copied %s to %s", final_target, resource_path)
            except OSError as copy_fail:
                BREAK_LOGGER.warning(
                    "Failed to copy %s to %s:\n  %s",
                    final_target,
                    resource_path,
                    copy_fail,
                )



"""Functionality for setting up the folder structure of both chests and shulker boxes"""

import logging
import re
from collections import Counter
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from urllib.parse import ParseResult

from pathvalidate import is_valid_filename

from . import filesystem as fs
from . import sync
from .enderchest import EnderChest, create_ender_chest
from .gather import gather_minecraft_instances
from .instance import InstanceSpec, normalize_modloader
from .inventory import (
    load_ender_chest,
    load_ender_chest_instances,
    load_ender_chest_remotes,
    load_shulker_boxes,
    render_instance,
    report_shulker_boxes,
)
from .loggers import CRAFT_LOGGER
from .prompt import NO, YES, confirm, prompt
from .remote import fetch_remotes_from_a_remote_ender_chest
from .shulker_box import ShulkerBox, create_shulker_box


def craft_ender_chest(
    minecraft_root: Path,
    copy_from: str | ParseResult | None = None,
    instance_search_paths: Iterable[str | Path] | None = None,
    remotes: (
        Iterable[str | ParseResult | tuple[str, str] | tuple[ParseResult, str]] | None
    ) = None,
    overwrite: bool = False,
) -> None:
    """Craft an EnderChest, either from the specified keyword arguments, or
    interactively via prompts

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff is in (or, at least, the
        one inside which you want to create your EnderChest)
    copy_from : URI, optional
        Optionally bootstrap your configuration by pulling the list of remotes
        from an existing remote EnderChest
    instance_search_paths : list of Paths, optional
        Any paths to search for Minecraft instances
    remotes : list of URIs or (URI, str) tuples, optional
        Any remotes you wish you manually specify. If used with `copy_from`, these
        will overwrite any remotes pulled from the remote EnderChest. When a
        (URI, str) tuple is provided, the second value will be used as the
        name/alias of the remote.
    overwrite : bool, optional
        This method will not overwrite an EnderChest instance installed within
        the `minecraft_root` unless the user provides `overwrite=True`

    Notes
    -----
    - The guided / interactive specifier will only be used if no other keyword
      arguments are provided (not even `overwrite=True`)
    - The instance searcher will first attempt to parse any instances it finds
      as official-launcher Minecrafts and then, if that doesn't work, will try
      parsing them as MultiMC-style instances.
    - The instance searcher is fully recursive, so keep that in mind before
      passing in, say "/"
    """
    if not minecraft_root.exists():
        CRAFT_LOGGER.error("The directory %s does not exist", minecraft_root)
        CRAFT_LOGGER.error("Aborting")
        return
    if (
        copy_from is None
        and instance_search_paths is None
        and remotes is None
        and not overwrite
    ):
        # then we go interactive
        try:
            ender_chest = specify_ender_chest_from_prompt(minecraft_root)
        except (FileExistsError, RuntimeError):
            CRAFT_LOGGER.error("Aborting")
            return
    else:
        try:
            fs.ender_chest_config(minecraft_root, check_exists=True)
            exist_message = (
                "There is already an EnderChest installed to %s",
                minecraft_root,
            )
            if overwrite:
                CRAFT_LOGGER.warning(*exist_message)
            else:
                CRAFT_LOGGER.error(*exist_message)
                CRAFT_LOGGER.error("Aborting")
                return
        except FileNotFoundError:
            pass  # no existing chest? no problem!

        ender_chest = EnderChest(minecraft_root)

        for search_path in instance_search_paths or ():
            for instance in gather_minecraft_instances(
                minecraft_root, Path(search_path), None
            ):
                ender_chest.register_instance(instance)

        if copy_from:
            try:
                for remote, alias in fetch_remotes_from_a_remote_ender_chest(copy_from):
                    if alias == ender_chest.name:
                        continue  # don't register yourself!
                    ender_chest.register_remote(remote, alias)
            except (RuntimeError, ValueError) as fetch_fail:
                CRAFT_LOGGER.error(
                    "Could not fetch remotes from %s:\n  %s", copy_from, fetch_fail
                )
                CRAFT_LOGGER.error("Aborting.")
                return

        for extra_remote in remotes or ():
            if isinstance(extra_remote, (str, ParseResult)):
                ender_chest.register_remote(extra_remote)
            else:
                ender_chest.register_remote(*extra_remote)

    create_ender_chest(minecraft_root, ender_chest)
    CRAFT_LOGGER.info(
        "\nNow craft some shulker boxes via\n$ enderchest craft shulker_box\n"
    )


def craft_shulker_box(
    minecraft_root: Path,
    name: str,
    priority: int | None = None,
    link_folders: Sequence[str] | None = None,
    instances: Sequence[str] | None = None,
    tags: Sequence[str] | None = None,
    hosts: Sequence[str] | None = None,
    overwrite: bool = False,
):
    """Craft a shulker box, either from the specified keyword arguments, or
    interactively via prompts

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    name : str
        A name to give to this shulker box
    priority : int, optional
        The priority for linking assets in the shulker box (higher priority
        shulkers are linked last)
    link_folders : list of str, optional
        The folders that should be linked in their entirety
    instances : list of str, optional
        The names of the instances you'd like to link to this shulker box
    tags : list of str, optional
        You can instead (see notes) provide a list of tags where any instances
        with those tags will be linked to this shulker box
    hosts : list of str, optional
        The EnderChest installations that this shulker box should be applied to
    overwrite : bool, optional
        This method will not overwrite an existing shulker box unless the user
        provides `overwrite=True`

    Notes
    -----
    - The guided / interactive specifier will only be used if no other keyword
      arguments are provided (not even `overwrite=True`)
    - The conditions specified by instances, tags and hosts are ANDed
      together--that is, if an instance is listed explicitly, but it doesn't
      match a provided tag, it will not link to this shulker box
    - Wildcards are supported for instances, tags and hosts (but not link-folders)
    - Not specifying instances, tags or hosts is equivalent to providing `["*"]`
    - When values are provided to the keyword arguments, no validation is performed
      to ensure that they are valid or actively in use
    """
    if not is_valid_filename(name):
        CRAFT_LOGGER.error("%s is not a valid name: must be usable as a filename", name)
        return

    try:
        folders = load_ender_chest(minecraft_root).shulker_box_folders
        if (
            priority is None
            and link_folders is None
            and instances is None
            and tags is None
            and hosts is None
            and not overwrite
        ):
            try:
                shulker_box = specify_shulker_box_from_prompt(minecraft_root, name)
            except FileExistsError as seat_taken:
                CRAFT_LOGGER.error(seat_taken)
                CRAFT_LOGGER.error("Aborting")
                return
        else:
            config_path = fs.shulker_box_config(minecraft_root, name)
            if config_path.exists():
                exist_message = (
                    "There is already a shulker box named %s in %s",
                    name,
                    fs.ender_chest_folder(minecraft_root),
                )
                if overwrite:
                    CRAFT_LOGGER.warning(*exist_message)
                else:
                    CRAFT_LOGGER.error(*exist_message)
                    CRAFT_LOGGER.error("Aborting")
                    return
            match_criteria: list[tuple[str, tuple[str, ...]]] = []
            if instances is not None:
                match_criteria.append(("instances", tuple(instances)))
            if tags is not None:
                match_criteria.append(("tags", tuple(tags)))
            if hosts is not None:
                match_criteria.append(("hosts", tuple(hosts)))
            shulker_box = ShulkerBox(
                priority=priority or 0,
                name=name,
                root=minecraft_root,
                match_criteria=tuple(match_criteria),
                link_folders=tuple(link_folders or ()),
            )
    except FileNotFoundError as no_ender_chest:
        CRAFT_LOGGER.error(no_ender_chest)
        return

    create_shulker_box(minecraft_root, shulker_box, folders)


def specify_ender_chest_from_prompt(minecraft_root: Path) -> EnderChest:
    """Parse an EnderChest based on interactive user input

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff is in (or, at least, the
        one inside which you want to create your EnderChest)

    Returns
    -------
    EnderChest
        The resulting EnderChest
    """
    try:
        root = fs.ender_chest_folder(minecraft_root)
        CRAFT_LOGGER.info(
            "This will overwrite the EnderChest configuration at %s.", root
        )
        if not confirm(default=False):
            message = f"Aborting: {fs.ender_chest_config(minecraft_root)} exists."
            raise FileExistsError(message)
    except FileNotFoundError:
        # good! Then we don't already have an EnderChest here
        CRAFT_LOGGER.debug("%s does not already contain an EnderChest", minecraft_root)

    instances: list[InstanceSpec] = []

    while True:
        search_home = prompt(
            "Would you like to search your home directory for the official launcher?",
            suggestion="Y/n",
        ).lower()
        if search_home == "" or search_home in YES:
            instances.extend(
                gather_minecraft_instances(minecraft_root, Path.home(), official=True)
            )
        elif search_home not in NO:
            continue
        break

    while True:
        search_here = prompt(
            "Would you like to search the current directory for MultiMC-type instances?",
            suggestion="Y/n",
        ).lower()
        if search_here == "" or search_here in YES:
            instances.extend(
                gather_minecraft_instances(minecraft_root, Path(), official=False)
            )
        elif search_here not in NO:
            continue
        break

    if minecraft_root.absolute() != Path().absolute():
        while True:
            search_mc_folder = prompt(
                f"Would you like to search {minecraft_root} for MultiMC-type instances?",
                suggestion="Y/n",
            ).lower()
            if search_mc_folder == "" or search_here in YES:
                instances.extend(
                    gather_minecraft_instances(
                        minecraft_root, minecraft_root, official=False
                    )
                )
            elif search_mc_folder not in NO:
                continue
            break

    CRAFT_LOGGER.info(
        "\nYou can always add more instances later using"
        "\n$ enderchest gather minecraft\n"
    )

    while True:
        remotes: list[tuple[ParseResult, str]] = []
        remote_uri = prompt(
            "Would you like to grab the list of remotes from another EnderChest?"
            "\nIf so, enter the URI of that EnderChest now (leave empty to skip)."
        )
        if remote_uri == "":
            break
        try:
            remotes.extend(fetch_remotes_from_a_remote_ender_chest(remote_uri))
        except Exception as fetch_fail:
            CRAFT_LOGGER.error(
                "Could not fetch remotes from %s\n  %s", remote_uri, fetch_fail
            )
            if not confirm(default=True):
                continue
        break

    CRAFT_LOGGER.info(
        "\nYou can always add more remotes later using"
        "\n$ enderchest gather enderchest\n"
    )

    while True:
        protocol = (
            prompt(
                (
                    "Specify the method for syncing with this EnderChest."
                    "\nSupported protocols are: " + ", ".join(sync.SUPPORTED_PROTOCOLS)
                ),
                suggestion=sync.DEFAULT_PROTOCOL,
            ).lower()
            or sync.DEFAULT_PROTOCOL
        )

        if protocol not in sync.SUPPORTED_PROTOCOLS:
            CRAFT_LOGGER.error("Unsupported protocol\n")
            continue
        break

    while True:
        default_netloc = sync.get_default_netloc()
        netloc = (
            prompt(
                (
                    "What's the address for accessing this machine?"
                    "\n(hostname or IP address, plus often a username)"
                ),
                suggestion=default_netloc,
            )
            or default_netloc
        )

        uri = ParseResult(
            scheme=protocol,
            netloc=netloc,
            path=minecraft_root.as_posix(),
            params="",
            query="",
            fragment="",
        )
        if not uri.hostname:
            CRAFT_LOGGER.error("Invalid hostname")
            continue
        break

    while True:
        name = (
            prompt("Provide a name for this EnderChest", suggestion=uri.hostname)
            or uri.hostname
        )
        if name in (alias for _, alias in remotes):
            CRAFT_LOGGER.error(
                "The name %s is already in use. Choose a different name.", name
            )
            continue
        break

    ender_chest = EnderChest(uri, name, remotes, instances)

    CRAFT_LOGGER.info(
        "\n%s\nPreparing to generate an EnderChest with the above configuration.",
        ender_chest.write_to_cfg(),
    )

    if not confirm(default=True):
        raise RuntimeError("EnderChest creation aborted.")

    return ender_chest


def specify_shulker_box_from_prompt(minecraft_root: Path, name: str) -> ShulkerBox:
    """Parse a shulker box based on interactive user input

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    name : str
        The name to give to the shulker box

    Returns
    -------
    ShulkerBox
        The resulting ShulkerBox
    """
    ender_chest = load_ender_chest(minecraft_root)
    shulker_root = fs.shulker_box_root(minecraft_root, name)
    if shulker_root in shulker_root.parent.iterdir():
        if not shulker_root.is_dir():
            raise FileExistsError(
                f"A file named {name} already exists in your EnderChest folder."
            )
        CRAFT_LOGGER.warning(
            "There is already a folder named %s in your EnderChest folder.", name
        )
        if not confirm(default=False):
            raise FileExistsError(
                f"There is already a folder named {name} in your EnderChest folder."
            )

    shulker_box = ShulkerBox(0, name, shulker_root, (), ())

    def refresh_ender_chest_instance_list() -> Sequence[InstanceSpec]:
        """The primary reason to lambda-fy this is to re-print the instance list."""
        return load_ender_chest_instances(minecraft_root)

    instances = refresh_ender_chest_instance_list()

    explicit_type = "name"
    if len(instances) > 0:
        explicit_type = "number"
    while True:
        selection_type = prompt(
            f"Would you like to specify instances by [F]ilter or by [N]{explicit_type[1:]}?"
        ).lower()
        match selection_type:
            case "f" | "filter":
                shulker_box = _prompt_for_filters(shulker_box, instances)
            case "n":
                if explicit_type == "name":
                    shulker_box = _prompt_for_instance_names(shulker_box)
                else:  # if explicit_type == "number"
                    shulker_box = _prompt_for_instance_numbers(
                        shulker_box, instances, refresh_ender_chest_instance_list
                    )
            case "name":
                # yeah, this is always available
                shulker_box = _prompt_for_instance_names(shulker_box)
            case "number":
                if explicit_type == "name":
                    continue
                shulker_box = _prompt_for_instance_numbers(
                    shulker_box, instances, refresh_ender_chest_instance_list
                )
            case _:
                continue
        break

    while True:
        selection_type = prompt(
            "Folders to Link?"
            "\nThe [G]lobal set is:"
            f' {", ".join(ender_chest.global_link_folders) or "(none)"}'
            "\nThe [S]tandard set is:"
            f' {", ".join(ender_chest.standard_link_folders) or "(none)"}'
            "\nYou can also choose [N]one or to [M]anually specify the folders to link",
            suggestion="S",
        ).lower()
        match selection_type:
            case "n" | "none":
                link_folders: tuple[str, ...] = ()
            case "g" | "global" | "global set":
                link_folders = tuple(ender_chest.global_link_folders)
            case "s" | "standard" | "standard set" | "":
                link_folders = tuple(ender_chest.standard_link_folders)
            case "m" | "manual" | "manually specify":
                folder_choices = prompt(
                    "Specify the folders to link using a comma-separated list"
                    " (wildcards are not allowed)"
                )
                link_folders = tuple(
                    folder.strip() for folder in folder_choices.split(",")
                )
            case _:
                continue
        break

    while True:
        # this is such a kludge
        existing_shulker_boxes = load_shulker_boxes(
            minecraft_root, log_level=logging.DEBUG
        )
        if existing_shulker_boxes:
            report_shulker_boxes(
                existing_shulker_boxes, logging.INFO, "the current EnderChest"
            )

        value = (
            prompt(
                (
                    "What priority value should be assigned to this shulker box?"
                    "\nhigher number = applied later"
                ),
                suggestion="0",
            )
            or "0"
        )
        try:
            priority = int(value)
        except ValueError:
            continue
        break

    while True:
        _ = load_ender_chest_remotes(minecraft_root)  # to display some log messages
        values = (
            prompt(
                (
                    "What hosts (EnderChest installations) should use this shulker box?"
                    "\nProvide a comma-separated list (wildcards are allowed)"
                    "\nand remember to include the name of this EnderChest"
                    f' ("{ender_chest.name}")'
                ),
                suggestion="*",
            )
            or "*"
        )
        hosts = tuple(host.strip() for host in values.split(","))

        host = ender_chest.name

        if not shulker_box._replace(match_criteria=(("hosts", hosts),)).matches_host(
            host
        ):
            CRAFT_LOGGER.warning(
                "This shulker box will not link to any instances on this machine"
            )
            if not confirm(default=False):
                continue
        break

    shulker_box = shulker_box._replace(
        priority=priority,
        match_criteria=shulker_box.match_criteria + (("hosts", hosts),),
        link_folders=link_folders,
    )

    CRAFT_LOGGER.info(
        "\n%sPreparing to generate a shulker box with the above configuration.",
        shulker_box.write_to_cfg(),
    )

    if not confirm(default=True):
        raise RuntimeError("Shulker box creation aborted.")

    return shulker_box


def _prompt_for_filters(
    shulker_box: ShulkerBox, instances: Sequence[InstanceSpec]
) -> ShulkerBox:
    """Prompt the user for a ShulkerBox spec by filters

    Parameters
    ----------
    shulker_box : ShulkerBox
        The starting ShulkerBox (with no match critera)
    instances : list of InstanceSpec
       The list of instances registered to the EnderChest

    Returns
    -------
    ShulkerBox
        The updated shulker box (with filter-based match criteria)

    Notes
    -----
    When instances is non-empty, the prompt will check in with the user at
    each step to make sure that the specified filters are filtering as intended.
    If that list is empty (or at the point that the user has filtered down to
    an empty list) then the user's gonna be flying blind.
    """

    def selected_instances(tester: ShulkerBox) -> list[InstanceSpec]:
        matching = [instance for instance in instances if tester.matches(instance)]
        CRAFT_LOGGER.info(
            "Specified filters match the instances:\n%s",
            "\n".join(
                [
                    f"  {i + 1}. {render_instance(instance)}"
                    for i, instance in enumerate(matching)
                ]
            ),
        )
        return matching

    def check_progress(
        new_condition: str, values: Iterable[str]
    ) -> tuple[ShulkerBox | None, list[InstanceSpec]]:
        """As we add new conditions, report to the user the list of instances
        that this shulker will match and confirm with them that they want
        to continue.

        Parameters
        ----------
        new_condition : str
            The type of condition being added
        values : list-like of str
           The values for the new condition

        Returns
        -------
        ShulkerBox or None
            If the user confirms that things look good, this will return the
            shulker box, updated with the new condition. Otherwise, the first
            returned value will be None
        list of str
            The names of the instances matching the updated set of filters

        Notes
        -----
        If instances is empty, then this method won't check in with the user
        and will just return the updated shulker box.
        """
        tester = shulker_box._replace(
            match_criteria=shulker_box.match_criteria
            + ((new_condition, tuple(values)),)
        )

        if len(instances) == 0:
            return tester, []

        matches = selected_instances(tester)

        default = True
        if len(matches) == 0:
            CRAFT_LOGGER.warning("Filters do not match any registered instance.")
            default = False
        return tester if confirm(default=default) else None, matches

    while True:
        version_spec = (
            prompt(
                (
                    "Minecraft versions:"
                    ' (e.g: "*", "1.19.1, 1.19.2, 1.19.3", "1.19.*", ">=1.19.0,<1.20")'
                ),
                suggestion="*",
            )
            or "*"
        )

        updated, matches = check_progress(
            "minecraft", (version.strip() for version in version_spec.split(", "))
        )
        if updated:
            shulker_box = updated
            instances = matches
            break

    while True:
        modloader = (
            prompt(
                (
                    "Modloader?"
                    "\n[N]one, For[G]e, Fa[B]ric, [Q]uilt, [L]iteLoader"
                    ' (or multiple, e.g: "B,Q", or any using "*")'
                ),
                suggestion="*",
            )
            or "*"
        )

        modloaders: set[str] = set()
        for entry in modloader.split(","):
            match entry.strip().lower():
                case "" | "n" | "none" | "vanilla":
                    modloaders.update(normalize_modloader(None))
                case "g" | "forge":
                    modloaders.update(normalize_modloader("Forge"))
                case "b" | "fabric" | "fabric loader":
                    modloaders.update(normalize_modloader("Fabric Loader"))
                case "q" | "quilt":
                    modloaders.update(normalize_modloader("Quilt Loader"))
                case "l" | "liteloader":
                    modloaders.update(normalize_modloader("LiteLoader"))
                case _:
                    modloaders.update(normalize_modloader(entry))

        updated, matches = check_progress("modloader", sorted(modloaders))
        if updated:
            shulker_box = updated
            instances = matches
            break

    while True:
        tag_count = Counter(sum((instance.tags for instance in instances), ()))
        # TODO: should this be most common among matches?
        example_tags: list[str] = [tag for tag, _ in tag_count.most_common(5)]
        CRAFT_LOGGER.debug(
            "Tag counts:\n%s",
            "\n".join(f"  - {tag}: {count}" for tag, count in tag_count.items()),
        )

        if len(example_tags) == 0:
            # provide examples if the user isn't using tags
            example_tags = [
                "vanilla-plus",
                "multiplayer",
                "modded",
                "dev",
                "april-fools",
            ]

        tags = (
            prompt(
                "Tags?"
                f'\ne.g.{", ".join(example_tags)}'
                "\n(or multiple using comma-separated lists or wildcards)"
                "\nNote: tag-matching is not case-sensitive."
                '\nNote: you can also specify a tag to exclude by prefacing it with "!"',
                suggestion="*",
            )
            or "*"
        )

        updated, matches = check_progress(
            "tags", (tag.strip() for tag in tags.split(","))
        )
        if updated:
            shulker_box = updated
            instances = matches
            break

    if len(instances) > 0:
        shulker_box = _prompt_for_instance_numbers(
            shulker_box,
            selected_instances(shulker_box),  # display the list again
            lambda: selected_instances(shulker_box),
            exclude=True,
        )

    return shulker_box


def _prompt_for_instance_names(shulker_box: ShulkerBox) -> ShulkerBox:
    """Prompt a user for the names of specific instances, then add that list
    to the shulker box spec.

    Parameters
    ----------
    shulker_box : ShulkerBox
        The starting ShulkerBox, presumably with limited or no match criteria

    Returns
    -------
    ShulkerBox
        The updated shulker box (with explicit criteria based on instance names)

    Notes
    -----
    This method does not validate against lists of known instances
    """
    instances = tuple(
        entry.strip()
        for entry in prompt(
            "Specify instances by name, separated by commas."
            "\nNote: this is case-sensitive."
            "\nNote: You can also use wildcards (? and *)"
            '\nNote: you can also specify an instance to exclude by prefacing it with "!"',
            suggestion="*",
        ).split(",")
    )
    if instances == ("",):
        instances = ("*",)

    if instances == ("*",):
        CRAFT_LOGGER.warning(
            "This shulker box will be applied to all instances,"
            " including ones you create in the future."
        )
        default = False
    else:
        CRAFT_LOGGER.info(
            "You specified the following instances:\n%s",
            "\n".join([f"  - {name}" for name in instances]),
        )
        default = True

    if not confirm(default=default):
        CRAFT_LOGGER.debug("Trying again to prompt for instance names")
        return _prompt_for_instance_names(shulker_box)

    return shulker_box._replace(
        match_criteria=shulker_box.match_criteria + (("instances", instances),)
    )


def _prompt_for_instance_numbers(
    shulker_box: ShulkerBox,
    instances: Sequence[InstanceSpec],
    instance_loader: Callable[[], Sequence[InstanceSpec]],
    exclude: bool = False,
) -> ShulkerBox:
    """Prompt the user to specify the instances they'd like by number

    Parameters
    ----------
    shulker_box : ShulkerBox
        The starting ShulkerBox, presumably with limited or no match criteria
    instances : list of InstanceSpec
       The list of instances registered to the EnderChest
    instance_loader : method that returns a list of InstanceSpec
        A method that when called, prints and returns a refreshed list of instances
        registered to the EnderChest
    exclude: bool, optional
        To modify this method to prompt for and return a list of shulker boxes to _exclude_
        instead of _include_, pass in `exclude=True`

    Returns
    -------
    ShulkerBox
        The updated shulker box (with explicit criteria based on instance names)
    """
    selections = prompt(
        (
            "Which instances would you like to {}?".format(
                "explicitly exclude" if exclude else "include"
            )
            + '\ne.g. "1,2,3", "1-3", "1-6"'
            + (' or "*" to specify all' if not exclude else "")
        ),
        suggestion="" if "exclude" else "*",
    )
    selections = re.sub("/s", " ", selections)  # normalize whitespace
    if selections == "":
        if exclude:
            return shulker_box
        selections = "*"

    if re.search("[^0-9-,* ]", selections):  # check for invalid characters
        CRAFT_LOGGER.error("Invalid selection.\n")
        return _prompt_for_instance_numbers(
            shulker_box, instance_loader(), instance_loader, exclude=exclude
        )

    selected_instances: set[str] = set()
    for entry in selections.split(","):
        match entry.replace(" ", ""):
            case "*":
                selected_instances.update(instance.name for instance in instances)
                break  # because it's not like there's any that can be added
            case value if value.isdigit():
                # luckily we don't need to worry about negative numbers
                index = int(value) - 1
                if index < 0 or index >= len(instances):
                    CRAFT_LOGGER.error("Invalid selection: %s is out of range\n", entry)
                    return _prompt_for_instance_numbers(
                        shulker_box, instance_loader(), instance_loader, exclude=exclude
                    )
                selected_instances.add(instances[index].name)
            case value if match := re.match("([0-9]+)-([0-9]+)$", value):
                bounds = tuple(int(bound) for bound in match.groups())
                if bounds[0] > bounds[1]:
                    CRAFT_LOGGER.error(
                        "Invalid selection: %s is not a valid range\n", entry
                    )
                    return _prompt_for_instance_numbers(
                        shulker_box, instance_loader(), instance_loader, exclude=exclude
                    )
                if max(bounds) > len(instances) or min(bounds) < 1:
                    CRAFT_LOGGER.error("Invalid selection: %s is out of range\n", entry)
                    return _prompt_for_instance_numbers(
                        shulker_box, instance_loader(), instance_loader, exclude=exclude
                    )
                selected_instances.update(
                    instance.name for instance in instances[bounds[0] - 1 : bounds[1]]
                )
            case _:
                CRAFT_LOGGER.error("Invalid selection.\n")
                return _prompt_for_instance_numbers(
                    shulker_box, instance_loader(), instance_loader, exclude=exclude
                )

    choices = tuple(
        instance.name for instance in instances if instance.name in selected_instances
    )
    if len(choices) == 0:  # this might not be possible to trigger
        return shulker_box

    CRAFT_LOGGER.info(
        "You selected to %s the instances:\n%s",
        "explicitly exclude" if exclude else "include",
        "\n".join([f"  - {name}" for name in choices]),
    )
    if not confirm(default=True):
        CRAFT_LOGGER.debug("Trying again to prompt for instance numbers")
        CRAFT_LOGGER.info("")  # just making a newline
        return _prompt_for_instance_numbers(
            shulker_box, instance_loader(), instance_loader, exclude=exclude
        )
    if exclude:
        choices = tuple(f"!{choice}" for choice in choices)

    prior_criteria: list[tuple[str, tuple[str, ...]]] = []
    prior_instance_selections: list[str] = []
    for condition, values in shulker_box.match_criteria:
        if condition == "instances":
            prior_instance_selections.extend(values)
        else:
            prior_criteria.append((condition, values))

    return shulker_box._replace(
        match_criteria=(
            *prior_criteria,
            ("instances", (*prior_instance_selections, *choices)),
        )
    )



"""shutil-based sync implementation"""

import fnmatch
import logging
import os
import shutil
import stat
from collections.abc import Callable, Collection
from pathlib import Path
from urllib.parse import ParseResult

from . import (
    SYNC_LOGGER,
    Op,
    abspath_from_uri,
    diff,
    filter_contents,
    generate_sync_report,
    is_identical,
)


def get_contents(path: Path) -> list[tuple[Path, os.stat_result]]:
    """Recursively list the contents of a local directory

    Parameters
    ----------
    path : Path
        The path to scan

    Returns
    -------
    list of filenames and attributes
        The attributes of all files, folders and symlinks found under the
        specified path

    Notes
    -----
    - This list will be sorted from shortest path to longest (so that parent
      directories come before their children)
    - The paths returned are all relative to the provided path
    """
    SYNC_LOGGER.debug("Getting contents of %s", path)
    return sorted(
        ((p.relative_to(path), p.lstat()) for p in path.rglob("**/*")),
        key=lambda x: len(str(x[0])),
    )


def copy(
    source_path: Path,
    destination_folder: Path,
    exclude: Collection[str],
    delete: bool,
    dry_run: bool,
) -> None:
    """Copy the specified source file or folder to the provided destination,
    overwriting any existing files and deleting any that weren't in the source

    Parameters
    ----------
    source_path : ParseResult
        The file or folder to copy
    destination_folder : Path
        The destination to put the source file(s)
    exclude : list of str
        Any patterns that should be excluded from the sync (and sync)
    delete : bool
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source.
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)

    Notes
    -----
    If the source file does not exist, the destination file will simply be deleted
    (if it exists)
    """

    ignore = ignore_patterns(*exclude)
    SYNC_LOGGER.debug("Ignoring patterns: %s", exclude)

    destination_path = destination_folder / source_path.name
    if destination_path.is_symlink() and not destination_path.is_dir():
        SYNC_LOGGER.warning("Removing symlink %s", destination_path)
        if not dry_run:
            destination_path.unlink()
        else:
            SYNC_LOGGER.debug("And replacing it entirely with %s", source_path)
            return
    elif destination_path.exists() and not destination_path.is_dir():
        SYNC_LOGGER.warning("Deleting file %s", destination_path)
        if not dry_run:
            destination_path.unlink()
        else:
            SYNC_LOGGER.debug("And replacing it entirely with %s", source_path)
            return
    else:
        if not dry_run:
            destination_folder.mkdir(parents=True, exist_ok=True)

    SYNC_LOGGER.debug("Copying %s into %s", source_path, destination_folder)

    if source_path.exists() and not source_path.is_dir():
        if destination_path.exists() and is_identical(
            source_path.stat(), destination_path.stat()
        ):
            SYNC_LOGGER.warning(
                "%s and %s are identical. No copy needed.",
                source_path,
                destination_path,
            )
            return
        SYNC_LOGGER.debug(
            "Copying file %s to %s",
            source_path,
            destination_path,
        )
        if not dry_run:
            shutil.copy2(source_path, destination_path, follow_symlinks=False)
        return

    source_contents = filter_contents(
        get_contents(source_path), exclude, prefix=source_path
    )
    destination_contents = filter_contents(
        get_contents(destination_path), exclude, prefix=destination_path
    )

    sync_diff = diff(source_contents, destination_contents)

    if dry_run:
        generate_sync_report(sync_diff)
        return

    for path, path_stat, operation in sync_diff:
        match (operation, stat.S_ISDIR(path_stat.st_mode or 0)):
            case (Op.CREATE, True):
                SYNC_LOGGER.debug("Creating directory %s", destination_path / path)
                (destination_path / path).mkdir(parents=True, exist_ok=True)
            case (Op.CREATE, False) | (Op.REPLACE, False):
                SYNC_LOGGER.debug(
                    "Copying file %s to %s",
                    source_path / path,
                    destination_path / path,
                )
                (destination_path / path).unlink(missing_ok=True)
                if (source_path / path).is_symlink():
                    (destination_path / path).symlink_to(
                        (source_path / path).readlink()
                    )
                else:
                    shutil.copy2(
                        source_path / path,
                        destination_path / path,
                        follow_symlinks=False,
                    )
            case (Op.REPLACE, True):
                # this would be replacing a file with a directory
                SYNC_LOGGER.debug("Deleting file %s", destination_path / path)
                (destination_path / path).unlink()
                SYNC_LOGGER.debug(
                    "Copying directory %s to %s",
                    source_path / path,
                    destination_path / path,
                )
                shutil.copytree(
                    source_path / path,
                    destination_path / path,
                    symlinks=True,
                    ignore=ignore,
                    dirs_exist_ok=True,
                )
            case (Op.DELETE, True):
                # recall that for deletions, it's the *destination's* stats
                if delete:
                    clean(destination_path / path, ignore, dry_run)
            case (Op.DELETE, False):
                if delete:
                    SYNC_LOGGER.debug("Deleting file %s", destination_path / path)
                    (destination_path / path).unlink()
            case op, is_dir:  # pragma: no cover
                raise NotImplementedError(
                    f"Don't know how to handle {op} of {'directory' if is_dir else 'file'}"
                )


def clean(
    root: Path,
    ignore: Callable[[str, Collection[str]], set[str]],
    dry_run: bool,
) -> None:
    """Recursively remove all files and symlinks from the root path while
    respecting the provided ignore pattern

    Parameters
    ----------
    root : Path
        The root directory. And this should absolutely be a directory.
    ignore : Callable
        The ignore pattern created by `ignore_pattern` that specifies
        which files to ignore.
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    """
    log_level = logging.INFO if dry_run else logging.DEBUG
    contents = list(root.iterdir())
    ignore_me = ignore(
        os.fspath(root),
        [path.name for path in contents],
    )

    for path in contents:
        if path.name in ignore_me:
            SYNC_LOGGER.debug("Skipping %s", path)
            continue
        if path.is_symlink():
            SYNC_LOGGER.log(log_level, "Removing symlink %s", path)
            if not dry_run:
                path.unlink()
        elif path.is_dir():
            clean(path, ignore, dry_run)
        else:
            SYNC_LOGGER.log(log_level, "Deleting %s", path)
            if not dry_run:
                path.unlink()

    # check if folder is now empty
    if not list(root.iterdir()):
        SYNC_LOGGER.log(log_level, "Removing empty %s", root)
        if not dry_run:
            root.rmdir()


def ignore_patterns(*patterns: str) -> Callable[[str, Collection[str]], set[str]]:
    """shutil.ignore_patterns doesn't support checking absolute paths,
    so we gotta roll our own.

    This implementation is adapted from
    https://github.com/python/cpython/blob/3.11/Lib/shutil.py#L440-L450 and
    https://stackoverflow.com/a/7842224

    Parameters
    ----------
    *patterns : str
        The patterns to match

    Returns
    -------
    Callable
        An "ignore" filter suitable for use in `shutil.copytree`
    """

    def _ignore_patterns(path: str, names: Collection[str]) -> set[str]:
        ignored_names: set[str] = set()
        for pattern in patterns:
            path_parts: list[str] = os.path.normpath(path).split(os.sep)
            pattern_depth = len(os.path.normpath(pattern).split(os.sep)) - 1
            if pattern_depth == 0:
                match_paths: Collection[str] = names
            else:
                match_paths = [
                    os.path.join(*path_parts[-pattern_depth:], name) for name in names
                ]
            ignored_names.update(
                os.path.split(match)[-1]
                for match in fnmatch.filter(match_paths, pattern)
            )
        return ignored_names

    return _ignore_patterns


def pull(
    remote_uri: ParseResult,
    local_path: Path,
    exclude: Collection[str],
    dry_run: bool,
    delete: bool = True,
    **unsupported_kwargs,
) -> None:
    """Copy an upstream file or folder into the specified location, where the remote
    is another folder on this machine. This will overwrite any files and folders
    already at the destination.

    Parameters
    ----------
    remote_uri : ParseResult
        The URI for the remote resource to copy from. See notes.
    local_path : Path
        The destination folder
    exclude : list of str
        Any patterns that should be excluded from the sync
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    delete : bool, optional
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source. Default is True.
    **unsupported_kwargs
        Any other provided options will be ignored

    Raises
    ------
    FileNotFoundError
        If the destination folder does not exist

    Notes
    -----
    - This method is only meant to be used for local files specified using
      the file:// protocol, but it does not perform any validation on the URI to
      ensure that the schema is correct or that the hostname corresponds to this
      machine. This method does not support user authentication
      (running the copy as a different user).
    - If the destination folder does not already exist, this method will not
      create it or its parent directories.
    """
    source_path = abspath_from_uri(remote_uri).expanduser()
    destination_folder = local_path

    if not destination_folder.exists():
        raise FileNotFoundError(f"{local_path} does not exist")
    if not source_path.exists():
        raise FileNotFoundError(f"{remote_uri.geturl()} does not exist")
    if unsupported_kwargs:
        SYNC_LOGGER.debug(
            "The following command-line options are ignored for this protocol:\n%s",
            "\n".join("  {}: {}".format(*item) for item in unsupported_kwargs.items()),
        )

    copy(source_path, destination_folder, exclude, delete=delete, dry_run=dry_run)


def push(
    local_path: Path,
    remote_uri: ParseResult,
    exclude: Collection[str],
    dry_run: bool,
    delete: bool = True,
    **unsupported_kwargs,
) -> None:
    """Copy a local file or folder into the specified location, where the remote
    is another folder on this machine. This will overwrite any files and folders
    already at the destination.

    Parameters
    ----------
    local_path : Path
        The file or folder to copy
    remote_uri : ParseResult
        The URI for the remote location to copy into. See notes.
    exclude : list of str
        Any patterns that should be excluded from the sync
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    delete : bool, optional
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source. Default is True.
    **unsupported_kwargs
        Any other provided options will be ignored

    Raises
    ------
    FileNotFoundError
        If the destination folder does not exist

    Notes
    -----
    - This method is only meant to be used for local files specified using
      the file:// protocol, but it does not perform any validation on the URI to
      ensure that the schema is correct or that the hostname corresponds to this
      machine. This method does not support user authentication
      (running the copy as a different user).
    - If the destination folder does not already exist, this method will not
      create it or its parent directories.
    """
    source_path = local_path
    destination_folder = abspath_from_uri(remote_uri).expanduser()

    if not destination_folder.exists():
        raise FileNotFoundError(f"{remote_uri.geturl()} does not exist")
    if not source_path.exists():
        raise FileNotFoundError(f"{source_path} does not exist")
    if unsupported_kwargs:
        SYNC_LOGGER.debug(
            "The following command-line options are ignored for this protocol:\n%s",
            "\n".join("  {}: {}".format(*item) for item in unsupported_kwargs.items()),
        )

    copy(source_path, destination_folder, exclude, delete=delete, dry_run=dry_run)



"""Useful setup / teardown fixtures"""

from collections.abc import Generator
from importlib.resources import as_file
from pathlib import Path

import pytest

from .testing_files import CLIENT_OPTIONS
from .utils import (
    TESTING_SHULKER_CONFIGS,
    populate_instances_folder,
    populate_official_minecraft_folder,
    pre_populate_enderchest,
)


@pytest.fixture
def use_local_ssh(request):
    """h/t https://stackoverflow.com/a/55003726
    (though it's also in the pytest docs lol)
    """
    return request.config.getoption("--use-local-ssh", False)


@pytest.fixture
def file_system(tmp_path) -> Generator[tuple[Path, Path], None, None]:
    """Create a testing file system and throw a bunch of random files across
    it.

    This method serves as both a setup and a teardown, as nothing
    in EnderChest's operations should be messing with these files, and
    this fixture asserts as such"""

    minecraft_root = tmp_path / "minecraft"
    minecraft_root.mkdir(parents=True)

    populate_instances_folder(minecraft_root)

    home = tmp_path / "home"
    home.mkdir(parents=True)

    populate_official_minecraft_folder(home / ".minecraft")

    (minecraft_root / "worlds").mkdir(
        parents=True
    )  # makes sense to store the actual saves outside the EnderChest folder
    (minecraft_root / "workspace").mkdir(
        parents=True
    )  # generic other stuff that exists in the minecraft root

    # files that EnderChest should not touch
    # (map) of file paths to their contents
    do_not_touch: dict[Path, str] = {
        minecraft_root
        / "README.md": "#Hello\n\nI'm your friendly neighborhood README!\n",
        (
            minecraft_root / "workspace" / "cool_project.py"
        ): 'print("Gosh, I hope no one deletes or overwrites me")\n',
        minecraft_root / "worlds" / "olam" / "level.dat": "level dis\n",
        minecraft_root / "worlds" / "testbench" / "level.dat": "hello world\n",
        (
            minecraft_root / "workspace" / "neat_resource_pack" / "pack.mcmeta"
        ): '{"so": "meta"}\n',
        (minecraft_root / "resourcepacks" / "TEAVSRP.zip"): "Breaking News!\n",
        (
            minecraft_root / "crash-reports" / "20230524.log"
        ): "ERROR: Everything is broken\nWARNING: And somehow also on fire\n",
        (
            minecraft_root / "instances" / "axolotl" / ".minecraft" / "options.txt"
        ): "renderDistance:1\n",
        home / "Pictures" / "Screenshots" / "sunrise.png": "PRETTY!!!\n",
    }

    with as_file(CLIENT_OPTIONS) as options_txt:
        do_not_touch[home / ".minecraft" / "options.txt"] = options_txt.read_text()

    mod_builds_folder = minecraft_root / "workspace" / "BestModEver" / "build" / "libs"

    do_not_touch.update(
        {
            mod_builds_folder / "BME_1.19_alpha.jar": "alfalfa",
            mod_builds_folder / "BME_1.19.1_beta.jar": "beater",
            mod_builds_folder / "BME_1.19.2_nightly.jar": "can i get a bat",
        }
    )

    chest_folder = minecraft_root / "EnderChest"

    do_not_touch.update(
        {
            chest_folder / ".git" / "log": "i committed some stuff\n",
            chest_folder / "global" / "usercache.json": "alexander\nmomoa\nbateman\n",
            (
                chest_folder / "global" / "config" / "iris.properties"
            ): "flower or aperture?\n",
        }
    )

    symlinks: dict[Path, Path] = {  # map of links to targets
        minecraft_root / "official_launcher" / ".minecraft": home / ".minecraft",
        (
            minecraft_root
            / "instances"
            / "chest-boat"
            / "minecraft"
            / "mods"
            / "BME.jar"
        ): (mod_builds_folder / "BME_1.19_alpha.jar"),
        (chest_folder / "global" / "crash-reports"): (minecraft_root / "crash-reports"),
        (chest_folder / "global" / "resourcepacks"): (minecraft_root / "resourcepacks"),
        (chest_folder / "global" / "screenshots"): (home / "Pictures" / "Screenshots"),
        (chest_folder / "global" / "saves" / "test"): (
            minecraft_root / "worlds" / "testbench"
        ),
        (chest_folder / "1.19" / "saves" / "olam"): (
            minecraft_root / "worlds" / "olam"
        ),
        (chest_folder / "optifine" / "mods" / "BME.jar"): (
            mod_builds_folder / "BME_1.19.2_nightly.jar"
        ),
    }

    for path, contents in do_not_touch.items():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(contents)

    for link_path, target in symlinks.items():
        link_path.parent.mkdir(parents=True, exist_ok=True)
        link_path.symlink_to(target, target_is_directory=target.is_dir())

    yield home, minecraft_root

    # check on teardown that all the symlinks still point to the right places
    for link_path, target in symlinks.items():
        assert link_path.resolve() == target.resolve()

    # check on teardown that all those "do_not_touch" files are untouched
    for path, contents in do_not_touch.items():
        assert not path.is_symlink()
        assert path.read_text() == contents


@pytest.fixture
def minecraft_root(file_system, monkeypatch):
    """Direct fixture pointing to the parent of the EnderChest folder and make
    sure all of our tests are running out of the minecraft root"""
    monkeypatch.chdir(file_system[1])
    yield file_system[1]


@pytest.fixture
def home(file_system, monkeypatch):
    """Direct fixture pointing to the user's "home" folder (with monkeypatching
    used to make the OS treat it as home)
    """
    home = file_system[0]

    # make sure SSH keys are still accessible post-fixting
    (home / ".ssh").symlink_to(Path.home() / ".ssh", target_is_directory=True)

    monkeypatch.setenv("HOME", str(home))  # posix
    monkeypatch.setenv("USERPROFILE", str(home))  # windows

    # check this right away so we can abort testing if it's not working
    assert Path.home() == home

    yield home


@pytest.fixture
def multi_box_setup_teardown(minecraft_root, home):
    """Setup / teardown for complex tests involving multiple boxes
    and overlapping resources"""
    chest_folder = minecraft_root / "EnderChest"
    pre_populate_enderchest(chest_folder, *TESTING_SHULKER_CONFIGS)

    do_not_touch = {
        (chest_folder / "global" / "resourcepacks" / "stuff.zip"): "dfgwhgsadfhsd",
        (chest_folder / "global" / "logs" / "bumpona.log"): (
            "Like a bump on a bump on a log, baby.\n"
            "Like I'm in a fist fight with a fog, baby.\n"
            "Step-ball-change and a pirouette.\n"
            "And I regret, I regret.\n"
        ),
        (chest_folder / "1.19" / "mods" / "FoxNap.jar"): "hello-maestro",
        (chest_folder / "1.19" / "options.txt"): "autoJump:true",
        (
            chest_folder / "vanilla" / "data" / "achievements.txt"
        ): "Spelled acheivements correctly!",
        (chest_folder / "optifine" / "mods" / "optifine.jar"): "sodium4life",
        (chest_folder / "optifine" / "shaderpacks" / "Seuss CitH.zip"): (
            "But those trees! Oh those trees! But those truffula trees!"
            "\nAll resplendent and gorgeous in ray-traced 3Ds"
        ),
        (
            chest_folder / "optifine" / "resourcepacks" / "stuff.zip"
        ): "optifine-optimized!",
        (
            minecraft_root
            / "instances"
            / "bee"
            / ".minecraft"
            / "shaderpacks"
            / "Seuss CitH.zip.txt"
        ): (
            "with settings at max"
            "\nits important to note"
            "\nthe lag is real bad"
            "\nbut just look at that goat!"
        ),
    }

    for path, contents in do_not_touch.items():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(contents)

    yield

    # check on teardown that all those "do_not_touch" files are untouched
    for path, contents in do_not_touch.items():
        assert path.read_text("utf-8") == contents


@pytest.fixture(autouse=True)
def set_log_levels(caplog):
    with caplog.at_level(20):
        yield



"""Specification of a Minecraft instance"""

import re
from configparser import SectionProxy
from pathlib import Path
from typing import NamedTuple

from . import config as cfg


class InstanceSpec(NamedTuple):
    """Specification of a Minecraft instance

    Parameters
    ----------
    name : str
        The "display name" for the instance
    root : Path
        The path to its ".minecraft" folder
    minecraft_versions : list-like of str
        The minecraft versions of this instance. This is typically a 1-tuple,
        but some loaders (such as the official one) will just comingle all
        your assets together across all profiles
    modloader : str
        The (display) name of the modloader (vanilla corresponds to "")
    tags : list-like of str
        The tags assigned to this instance, including both the ones assigned
        in the launcher (groups) and the ones assigned by hand.
    """

    name: str
    root: Path
    minecraft_versions: tuple[str, ...]
    modloader: str
    groups_: tuple[str, ...]
    tags_: tuple[str, ...]

    @classmethod
    def from_cfg(cls, section: SectionProxy) -> "InstanceSpec":
        """Parse an instance spec as read in from the enderchest config file

        Parameters
        ----------
        section : dict-like of str to str
            The section in the enderchest config as parsed by a ConfigParser

        Returns
        -------
        InstanceSpec
            The resulting InstanceSpec

        Raises
        ------
        KeyError
            If a required key is absent
        ValueError
            If a required entry cannot be parsed
        """
        return cls(
            section.name,
            Path(section["root"]),
            tuple(
                parse_version(version.strip())
                for version in cfg.parse_ini_list(
                    section.get(
                        "minecraft-version", section.get("minecraft_version", "")
                    )
                )
            ),
            normalize_modloader(section.get("modloader", None))[0],
            tuple(cfg.parse_ini_list(section.get("groups", ""))),
            tuple(cfg.parse_ini_list(section.get("tags", ""))),
        )

    @property
    def tags(self):
        return tuple(sorted({*self.groups_, *self.tags_}))


def normalize_modloader(loader: str | None) -> list[str]:
    """Implement common modloader aliases

    Parameters
    ----------
    loader : str
        User-provided modloader name

    Returns
    -------
    list of str
        The modloader values that should be checked against to match the user's
        intent
    """
    if loader is None:  # this would be from the instance spec
        return [""]
    match loader.lower().replace(" ", "").replace("-", "").replace("_", "").replace(
        "/", ""
    ):
        case "none" | "vanilla" | "minecraftserver":
            return [""]
        case "fabric" | "fabricloader":
            return ["Fabric Loader"]
        case "quilt" | "quiltloader":
            return ["Quilt Loader"]
        case "fabricquilt" | "quiltfabric" | "fabriclike" | "fabriccompatible":
            return ["Fabric Loader", "Quilt Loader"]
        case "forge" | "forgeloader" | "minecraftforge":
            return ["Forge"]
        case _:
            return [loader.title()]


def equals(
    minecraft_root: Path, instance: InstanceSpec, other_instance: InstanceSpec
) -> bool:
    """Determine whether two instances point to the same location

    Parameters
    ----------
    minecraft_root : Path
        The starting location (the parent of where your EnderChest folder lives)
    instance : InstanceSpec
        the first instance
    other_instance : InstanceSpec
        the instance to compare it against

    Returns
    -------
    bool
        True if and only if the two instances have the same root, with regards
        to the provided `minecraft_root`
    """
    path = minecraft_root / instance.root.expanduser()
    other_path = minecraft_root / other_instance.root.expanduser()
    return path.expanduser().resolve() == other_path.expanduser().resolve()


def parse_version(version_string: str) -> str:
    """The first release of each major Minecraft version doesn't follow strict
    major.minor.patch semver. This method appends the ".0" so that our version
    matcher doesn't mess up

    Parameters
    ----------
    version_string : str
        The version read in from the Minecraft instance's config

    Returns
    -------
    str
        Either the original version string or the original version string with
        ".0" tacked onto the end of it

    Notes
    -----
    Regex adapted straight from https://semver.org
    """
    if re.match(r"^(0|[1-9]\d*)\.(0|[1-9]\d*)$", version_string):
        return version_string + ".0"
    return version_string


def merge(*instances: InstanceSpec) -> InstanceSpec:
    """Merge multiple instances, layering information from the ones provided later
    on top of the ones provided earlier

    Parameters
    ----------
    *instances : InstanceSpec
        The instances to combine

    Returns
    -------
    InstanceSpec
        The merged instance

    Notes
    -----
    The resulting merged instance will use:
      - the first instance's name
      - the union of all non-group tags
      - all other data from the last instance
    """
    try:
        combined_instance = instances[-1]
    except IndexError as nothing_to_merge:
        raise ValueError(
            "Must provide at least one instance to merge"
        ) from nothing_to_merge
    tags = tuple(sorted(set(sum((instance.tags_ for instance in instances), ()))))
    return combined_instance._replace(name=instances[0].name, tags_=tags)



from pathlib import Path

import pytest

from enderchest import instance as i

from .utils import instance


class TestInstanceEquality:
    def test_two_instances_with_different_paths_are_not_equal(self):
        assert not i.equals(
            Path(), instance("foo", Path("a")), instance("foo", Path("b"))
        )

    def test_two_instances_with_the_same_path_are_equal(self):
        assert i.equals(Path(), instance("foo", Path("a")), instance("fufu", Path("a")))

    def test_two_instances_with_the_same_abspath_are_equal(self, minecraft_root):
        assert i.equals(
            minecraft_root,
            instance("foo", minecraft_root / "instances" / "axolotl"),
            instance("fufu", Path("instances/axolotl")),
        )

    def test_equality_check_works_with_expanduser(self, minecraft_root, home):
        assert i.equals(
            minecraft_root,
            instance("foo", home / "Library" / "SomeKindOfMC" / "Instances" / "foo"),
            instance("fufu", Path("~/Library/SomeKindOfMC/Instances/foo")),
        )

    def test_equality_check_works_with_minecraft_root_in_home(self, home):
        assert i.equals(
            Path("~/Library/SomeKindOfMC"),
            instance("foo", Path("Instances") / "foo"),
            instance("fufu", Path("~/Library/SomeKindOfMC/Instances/foo")),
        )

    def test_equality_resolves_symlinks(self, home, minecraft_root):
        assert i.equals(
            minecraft_root,
            instance("foo", Path("official_launcher") / ".minecraft"),
            instance("fufu", Path("~/.minecraft")),
        )


class TestInstanceMerging:
    def test_supplying_no_instances_raises_a_sensible_error(self):
        with pytest.raises(ValueError, match="at least one instance"):
            i.merge()

    def test_merging_a_single_instance_results_in_a_copy(self):
        original = instance(
            name="an instance",
            root=Path("."),
            minecraft_versions=("1.21.0",),
            modloader="linen",
            groups=("experimental",),
            tags=("a", "b"),
        )

        merged = i.merge(original)

        assert original is not merged
        assert original == merged

    @pytest.mark.parametrize(
        "field",
        (field for field in i.InstanceSpec._fields if field not in ("name", "tags_")),
    )
    def test_merged_fields_mostly_match_the_last_instance(self, field):
        instances = [
            instance("one", Path("path"), minecraft_versions=("1.20.2",)),
            instance(
                "two",
                Path("path"),
                modloader="blacksmith",
                groups=("forgery",),
                tags=("you're it",),
            ),
            instance(
                "three",
                Path("path"),
                minecraft_versions=("1.21-pre1",),
                groups=("next_gen",),
            ),
        ]

        merged = i.merge(*instances)

        assert getattr(merged, field) == getattr(instances[-1], field)

    def test_merged_name_matches_the_first_instance(self):
        instances = [
            instance("one", Path("path"), tags=("you're it",)),
            instance("two", Path("path"), tags=("hello", "friend")),
            instance("three", Path("path"), tags=("hello", "world")),
            instance("four", Path("path"), tags=("best", "friend")),
            instance("five", Path("path")),
        ]

        merged = i.merge(*instances)

        assert merged.name == "one"

    def test_merged_tags_are_the_union_of_all_instance_tags(self):
        instances = [
            instance("one", Path("path"), tags=("you're it",)),
            instance("two", Path("path"), tags=("hello", "friend")),
            instance("three", Path("path"), tags=("hello", "world")),
            instance("four", Path("path"), tags=("best", "friend")),
            instance("five", Path("path")),
        ]

        merged = i.merge(*instances)

        assert merged.tags_ == ("best", "friend", "hello", "world", "you're it")



"""A mock Paramiko SFTP client to be used for testing on systems without local SSH"""

import json
import os
import shutil
from collections.abc import Callable, Generator
from contextlib import contextmanager
from importlib.resources import as_file
from pathlib import Path
from typing import NamedTuple
from urllib.parse import ParseResult
from urllib.request import url2pathname

from .testing_files import LSTAT_CACHE


class CachedStat(NamedTuple):
    """Stat-like loaded from a file cache"""

    filename: str
    st_mode: int
    st_size: float
    st_atime: float
    st_mtime: float


A_DIRECTORY = CachedStat(
    filename="", st_mode=16877, st_size=-1, st_atime=-1, st_mtime=-1
)
A_FILE = CachedStat(filename="", st_mode=33188, st_size=-1, st_atime=-1, st_mtime=-1)
A_SYMLINK = CachedStat(filename="", st_mode=41471, st_size=-1, st_atime=-1, st_mtime=-1)


class MockSFTP:
    """Create a mock SFTP client suitable for testing

    Parameters
    ----------
    root : Path
        The top-most directory that will be accessed by this client (this
        corresponds to the path in the remote URI for the EnderChest)
    """

    def __init__(self, root: Path):
        with as_file(LSTAT_CACHE) as lstat_cache_file:
            cached_lstats: list[dict] = json.loads(lstat_cache_file.read_text("UTF-8"))

        self.lstat_cache: dict[Path, CachedStat] = {
            root
            / stat["filename"]: CachedStat(**stat)._replace(
                st_size=(root / stat["filename"]).stat().st_size,
                st_mtime=(root / stat["filename"]).stat().st_mtime,
                st_atime=(root / stat["filename"]).stat().st_atime,
            )
            for stat in cached_lstats
        }

    def lstat(self, path: str) -> CachedStat:
        """Return the cached file attributes for the specified path"""
        try:
            return self.lstat_cache[Path(url2pathname(path))]
        except KeyError as not_in_cache:
            # In a few places we get an lstat on stuff that's outside of the
            # cache. In those scenarios, we want to return CacheStats matching
            # the type of the file
            if Path(url2pathname(path)).is_symlink():
                return A_SYMLINK
            if Path(url2pathname(path)).is_dir():
                return A_DIRECTORY
            if Path(url2pathname(path)).exists():
                return A_FILE
            raise FileNotFoundError from not_in_cache

    def mkdir(self, path: str) -> None:
        """Make a directory on the "remote" file system"""
        Path(url2pathname(path)).mkdir()

    def symlink(self, target: str, path: str) -> None:
        """Create a symlink at the path pointing to the target"""
        Path(url2pathname(path)).symlink_to(Path(target))

    def readlink(self, path: str) -> str:
        """Get the target of the "remote" symlink"""
        return Path(url2pathname(path)).readlink().as_posix()

    def get(self, path: str, destination: Path) -> None:
        """ "Download" the "remote" file to the specified destination"""
        shutil.copy2(
            Path(url2pathname(path)),
            destination,
            follow_symlinks=False,
        )

    def put(self, source: Path, path: str) -> None:
        """ "Upload" the "remote" file to the specified destination"""
        shutil.copy2(
            source,
            Path(url2pathname(path)),
            follow_symlinks=False,
        )

    def utime(self, path: str, times: tuple[float, float]) -> None:
        """Set the modification and access times of a remote file"""
        os.utime(Path(url2pathname(path)), times)

    def remove(self, path: str) -> None:
        """Delete a file on the "remote" file system"""
        Path(url2pathname(path)).unlink()

    def rmdir(self, path: str) -> None:
        """Delete an empty folder on the "remote" file system"""
        Path(url2pathname(path)).rmdir()


def generate_mock_connect(mock_sftp: MockSFTP) -> Callable:
    """Create a method suitable to be used as a monkeypatch for `sync.sftp.connect`
    that instead yields the provided MockSFTP instance

    Parameters
    ----------
    mock_sftp : MockSFTP
        The mock SFTP client the method should connect to instead

    Returns
    -------
    Method
        A method suitable for use in monkeypatching `sync.sftp.connect`
    """

    @contextmanager
    def connect(
        uri: ParseResult, timeout: float | None = None
    ) -> Generator[MockSFTP, None, None]:
        yield mock_sftp

    return connect


def mock_rglob(client: MockSFTP, path: str) -> list[tuple[Path, CachedStat]]:
    """A method that simulates `sync.sftp.rglob` by relying on the
    `lstat_cache` of the client instead of actually recursively getting the
    file attributes of the given path

    Parameters
    ----------
    client : MockSFTP
        The mock SFTP client
    path : str
        This parameter is just used to determine which test is being run

    Returns
    -------
    list of (Path, SFTPAttributes-like) tuples
        The cached file attributes
    """
    if "somewhere_else" in Path(url2pathname(path)).parts:
        return []
    return list(client.lstat_cache.items())



"""Utilities for helping build interactive prompts"""

import getpass

CURSOR = "\x1b[35;1m==>\x1b[0m"

# https://stackoverflow.com/a/18472142
YES = ("y", "yes", "t", "true", "on", "1")

NO = ("n", "no", "f", "false", "off", "0")


def prompt(
    message: str, suggestion: str | None = None, is_password: bool = False
) -> str:
    """Prompt the user and return the response

    Parameters
    ----------
    message : str
        The prompt message
    suggestion : str, optional
        A suggested input. If None is provided, no suggestion will be shown.
    is_password : bool, optional
        If this is a prompt for a password, pass in `is_password = True` to
        hide the user's input

    Returns
    -------
    str
        The user-provided response

    Notes
    -----
    - The output will be stripped of trailing and leading whitespace, but no
      other validation or processing will be used.
    - Regardless of whether a suggestion is provided, if the user provides an
      empty input, this method will return an empty string. To reiterate: the
      suggestion *does not serve* as a default / fallback value.
    """
    lines = message.splitlines() + [""]
    message = "\n".join(f"{CURSOR}\x1b[1m {line}\x1b[0m" for line in lines)
    if suggestion is not None:
        message += f"\x1b[35;1m[{suggestion}]\x1b[0m "
    if is_password:
        return getpass.getpass(prompt=message)
    return input(message)


def confirm(default: bool) -> bool:
    """Confirm that the user wishes to continue

    Parameters
    ----------
    default : bool
        Whether the default selection should be True (yes) or False (no)

    Returns
    -------
    bool
        Whether the user has opted to continue
    """

    response = prompt("Do you wish to continue?", "Y/n" if default else "y/N").lower()

    if response == "":
        return default

    if response in YES:
        return True

    return False



"""paramiko-based sftp sync implementation"""

import os
import posixpath
import stat
from collections.abc import Collection, Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from urllib.parse import ParseResult, unquote
from urllib.request import url2pathname

import paramiko

from ..prompt import prompt
from . import (
    SYNC_LOGGER,
    Op,
    diff,
    file,
    filter_contents,
    generate_sync_report,
    is_identical,
)


@contextmanager
def connect(
    uri: ParseResult, timeout: float | None = None
) -> Generator[paramiko.sftp_client.SFTPClient, None, None]:
    """Yield an SFTPClient connected to the server specified by the given URI

    Parameters
    ----------
    uri : ParseResult
        The URI of the EnderChest to connect to
    timeout : float, optional
        The number of seconds to wait before timing out the sync operation.
        If None is provided, no explicit timeout value will be set.

    Yields
    ------
    SFTPClient
        A Paramiko SFTP client connected to the specified server

    Raises
    ------
    ValueError
        If the URI is invalid or the credentials are incorrect
    RuntimeError
        If the server cannot be reached
    """
    ssh_client = paramiko.client.SSHClient()
    ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    extra_kwargs: dict[str, Any] = {}
    if timeout is not None:
        extra_kwargs["timeout"] = timeout

    try:
        ssh_client.connect(
            uri.hostname or "localhost",
            port=uri.port or 22,
            username=uri.username,
            # note: passing in password is explicitly unsupported
            **extra_kwargs,
        )
    except paramiko.AuthenticationException:
        target = ((uri.username + "@") if uri.username else "") + (
            uri.hostname or "localhost"
        )

        SYNC_LOGGER.warning(
            "This machine is not set up for passwordless login to %s"
            "\nFor instructions on setting up public key-based authentication,"
            " which is both"
            "\nmore convenient and more secure, see:"
            "\nhttps://openbagtwo.github.io/EnderChest"
            "/dev/suggestions/#passwordless-ssh-authentication",
            target,
        )
        password = prompt(f"Please enter the password for {target}", is_password=True)
        try:
            ssh_client.connect(
                uri.hostname or "localhost",
                port=uri.port or 22,
                username=uri.username,
                password=password,
                **extra_kwargs,
            )
        except paramiko.AuthenticationException as bad_login:
            raise ValueError(
                "Authentication failed."
                " Did you supply the correct username and password?"
            ) from bad_login

    try:
        sftp_client = ssh_client.open_sftp()
        yield sftp_client
        sftp_client.close()
    finally:
        ssh_client.close()


def download_file(
    client: paramiko.sftp_client.SFTPClient,
    remote_loc: str,
    local_path: Path,
    remote_stat: paramiko.SFTPAttributes,
) -> None:
    """Download a file from a remote SFTP server and save it at the specified
    location.

    Parameters
    ----------
    client : Paramiko SFTP client
        An authenticated client connected to the remote server
    remote_loc : str
        The POSIX path of the file to download
    local_path : Path
        The path to locally save the file
    remote_stat : stat-like
        The `os.stat_result`-like properties of the remote object

    Notes
    -----
    This is a wrapper around `client.get()` that can handle symlinks and
    updating timestamps. It does not check if either path is valid, points
    to a file, lives in an existing folder, etc.
    """
    if stat.S_ISLNK(remote_stat.st_mode or 0):
        local_path.symlink_to(Path((client.readlink(remote_loc) or "")))
    else:
        client.get(remote_loc, local_path)
        if remote_stat.st_atime and remote_stat.st_mtime:
            os.utime(
                local_path,
                times=(remote_stat.st_atime, remote_stat.st_mtime),
            )


def upload_file(
    client: paramiko.sftp_client.SFTPClient,
    local_path: Path,
    remote_loc: str,
) -> None:
    """Upload a local file to a remote SFTP server

    Parameters
    ----------
    client : Paramiko SFTP client
        An authenticated client connected to the remote server
    local_path : Path
        The path of the file to upload
    remote_loc : str
        The POSIX path for the remote location to save the file

    Notes
    -----
    This is just a wrapper around `client.put()` that can handle symlinks.
    It does not check if either path is valid, points to a file, lives in an
    existing folder, etc.
    """
    if local_path.is_symlink():
        client.symlink(local_path.readlink().as_posix(), remote_loc)
    else:
        client.put(local_path, remote_loc)
        client.utime(
            remote_loc, times=(local_path.stat().st_atime, local_path.stat().st_mtime)
        )


def rglob(
    client: paramiko.sftp_client.SFTPClient, path: str
) -> list[tuple[Path, paramiko.sftp_attr.SFTPAttributes]]:
    """Recursively enumerate the contents of a remote directory

    Parameters
    ----------
    client : Paramiko SFTP client
        An authenticated client connected to the remote server
    path : str
        The absolute path to scan

    Returns
    -------
    list of (Path, SFTPAttributes) tuples
        The attributes of all files, folders and symlinks found under the
        specified path

    Notes
    -----
    - The paths returned are *absolute*
    - The search is performed depth-first
    """
    SYNC_LOGGER.debug("ls %s", path)
    top_level = client.listdir_attr(path)
    contents: list[tuple[Path, paramiko.sftp_attr.SFTPAttributes]] = []
    for remote_object in top_level:
        remote_object.filename = posixpath.join(path, remote_object.filename)
        contents.append((Path(url2pathname(remote_object.filename)), remote_object))
        if stat.S_ISDIR(remote_object.st_mode or 0):
            contents.extend(rglob(client, remote_object.filename))
    return contents


def get_contents(
    client: paramiko.sftp_client.SFTPClient, path: str
) -> list[tuple[Path, paramiko.sftp_attr.SFTPAttributes]]:
    """Recursively fetch the contents of a remote directory

    Parameters
    ----------
    client : Paramiko SFTP client
        An authenticated client connected to the remote server
    path : str
        The absolute path to scan

    Returns
    -------
    list of (Path, SFTPAttributes) tuples
        The attributes of all files, folders and symlinks found under the
        specified path

    Notes
    -----
    - This list is generated via a depth-first search so that all parent
      directories appear before their children
    - The paths returned are relative to the provided path
    """
    return [
        (p.relative_to(url2pathname(path)), path_stat)
        for p, path_stat in rglob(client, path)
    ]


def pull(
    remote_uri: ParseResult,
    local_path: Path,
    exclude: Collection[str],
    dry_run: bool,
    timeout: float | None = None,
    delete: bool = True,
    **unsupported_kwargs,
) -> None:
    """Sync an upstream file or folder into the specified location SFTP.
    This will overwrite any files and folders already at the destination.

    Parameters
    ----------
    remote_uri : ParseResult
        The URI for the remote resource to copy from
    local_path : Path
        The destination folder
    exclude : list of str
        Any patterns that should be excluded from the sync
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    timeout : float, optional
        The number of seconds to wait before timing out the sync operation.
        If None is provided, no explicit timeout value will be set.
    delete : bool
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source. Default is True.
    **unsupported_kwargs
        Any other provided options will be ignored

    Raises
    ------
    FileNotFoundError
        If the destination folder does not exist, or if the remote path
        does not exist
    OSError
        If the remote path cannot be accessed for any other reason (permissions,
        most likely)

    Notes
    -----
    - If the destination folder does not already exist, this method will not
      create it or its parent directories.
    """
    if not local_path.exists():
        raise FileNotFoundError(f"{local_path} does not exist")
    if unsupported_kwargs:
        SYNC_LOGGER.debug(
            "The following command-line options are ignored for this protocol:\n%s",
            "\n".join("  {}: {}".format(*item) for item in unsupported_kwargs.items()),
        )

    remote_loc = posixpath.normpath(unquote(remote_uri.path))
    destination_path = local_path / posixpath.basename(remote_loc)

    if destination_path.is_symlink() and not destination_path.is_dir():
        SYNC_LOGGER.warning("Removing symlink %s", destination_path)
        if not dry_run:
            destination_path.unlink()
        else:
            SYNC_LOGGER.debug(
                "And replacing it entirely with the remote's %s", remote_loc
            )
            return
    elif destination_path.exists() and not destination_path.is_dir():
        SYNC_LOGGER.warning("Deleting file %s", destination_path)
        if not dry_run:
            destination_path.unlink()
        else:
            SYNC_LOGGER.debug(
                "And replacing it entirely with the remote's %s", remote_loc
            )
            return

    with connect(uri=remote_uri, timeout=timeout) as remote:
        try:
            source_target = remote.lstat(remote_loc)
        except OSError as bad_target:
            raise type(bad_target)(
                f"Could not access {remote_loc} on remote: {bad_target}"
            )
        if not stat.S_ISDIR(source_target.st_mode or 0):
            if destination_path.exists() and is_identical(
                source_target, destination_path.stat()
            ):
                SYNC_LOGGER.warning(
                    "Remote file matches %s. No transfer needed.",
                    destination_path,
                )
                return
            SYNC_LOGGER.debug(
                "Downloading file %s from remote",
                destination_path,
            )
            if not dry_run:
                download_file(
                    remote,
                    remote_loc,
                    destination_path,
                    source_target,
                )
            return

        if not destination_path.exists():
            SYNC_LOGGER.debug(
                "Downloading the entire contents of the remote's %s", remote_loc
            )
            if dry_run:
                return
            destination_path.mkdir()

        source_contents = filter_contents(
            get_contents(remote, remote_loc),
            exclude,
            prefix=remote_loc,
        )
        destination_contents = filter_contents(
            file.get_contents(destination_path),
            exclude,
            prefix=destination_path,
        )

        sync_diff = diff(source_contents, destination_contents)

        if dry_run:
            generate_sync_report(sync_diff)
            return

        ignore = file.ignore_patterns(*exclude)
        for path, path_stat, operation in sync_diff:
            match (operation, stat.S_ISDIR(path_stat.st_mode or 0)):
                case (Op.CREATE, True):
                    SYNC_LOGGER.debug("Creating directory %s", destination_path / path)
                    (destination_path / path).mkdir(parents=True, exist_ok=True)
                case (Op.CREATE, False) | (Op.REPLACE, False):
                    SYNC_LOGGER.debug(
                        "Downloading file %s from remote",
                        destination_path / path,
                    )
                    (destination_path / path).unlink(missing_ok=True)
                    download_file(
                        remote,
                        posixpath.join(remote_loc, path.as_posix()),
                        destination_path / path,
                        path_stat,  # type: ignore[arg-type]
                    )
                case (Op.DELETE, True):
                    # recall that for deletions, it's the *destination's* stats
                    if delete:
                        file.clean(destination_path / path, ignore, dry_run)
                case (Op.DELETE, False):
                    SYNC_LOGGER.debug("Deleting file %s", destination_path / path)
                    if delete:
                        (destination_path / path).unlink()
                case op, is_dir:  # pragma: no cover
                    raise NotImplementedError(
                        f"Don't know how to handle {op} of {'directory' if is_dir else 'file'}"
                    )


def push(
    local_path: Path,
    remote_uri: ParseResult,
    exclude: Collection[str],
    dry_run: bool,
    timeout: float | None = None,
    delete: bool = True,
    **unsupported_kwargs,
) -> None:
    """Sync a local file or folder into the specified location using SFTP.
    This will overwrite any files and folders already at the destination.

    Parameters
    ----------
    local_path : Path
        The file or folder to copy
    remote_uri : ParseResult
        The URI for the remote location to copy into
    exclude : list of str
        Any patterns that should be excluded from the sync
    dry_run : bool
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them)
    timeout : float, optional
        The number of seconds to wait before timing out the sync operation.
        If None is provided, no explicit timeout value will be set.
    delete : bool, optional
        Whether part of the syncing should include deleting files at the destination
        that aren't at the source. Default is True.
    **unsupported_kwargs
        Any other provided options will be ignored

    Notes
    -----
    - If the destination folder does not already exist, this method will not
      create it or its parent directories.
    """
    if not local_path.exists():
        raise FileNotFoundError(f"{local_path} does not exist.")
    if unsupported_kwargs:
        SYNC_LOGGER.debug(
            "The following command-line options are ignored for this protocol:\n%s",
            "\n".join("  {}: {}".format(*item) for item in unsupported_kwargs.items()),
        )

    remote_parent = posixpath.normpath(unquote(remote_uri.path))

    with connect(uri=remote_uri, timeout=timeout) as remote:
        try:
            remote_folder_stat = remote.lstat(remote_parent)
        except OSError as bad_target:
            raise type(bad_target)(
                f"Could not access {remote_parent} on remote: {bad_target}"
            )
        if not stat.S_ISDIR(remote_folder_stat.st_mode or 0):
            raise NotADirectoryError(f"{remote_parent} on remote is not a directory.")

        remote_loc = posixpath.join(remote_parent, local_path.name)
        try:
            target_stat = remote.lstat(remote_loc)
        except FileNotFoundError:
            target_stat = None
        if not stat.S_ISDIR(local_path.stat().st_mode or 0):
            if target_stat and is_identical(local_path.stat(), target_stat):
                SYNC_LOGGER.warning("Remote file matches %s", local_path)
                return

            SYNC_LOGGER.debug(
                "Uploading file %s to remote",
                local_path,
            )
            if not dry_run:
                upload_file(remote, local_path, remote_loc)
            return
        if not target_stat:
            SYNC_LOGGER.debug("Uploading the entire contents %s", local_path)
            if dry_run:
                return
            remote.mkdir(remote_loc)
        elif not stat.S_ISDIR(target_stat.st_mode or 0):
            SYNC_LOGGER.warning(
                "Deleting remote file or symlink %s",
                remote_loc,
            )
            if dry_run:
                SYNC_LOGGER.debug("And replacing it entirely with %s", local_path)
                return
            remote.remove(remote_loc)
            remote.mkdir(remote_loc)

        source_contents = filter_contents(
            file.get_contents(local_path), exclude, prefix=local_path
        )
        destination_contents = filter_contents(
            get_contents(remote, remote_loc),
            exclude,
            prefix=remote_loc,
        )

        sync_diff = diff(source_contents, destination_contents)

        if dry_run:
            generate_sync_report(sync_diff)
            return

        for path, path_stat, operation in sync_diff:
            posix_path = posixpath.join(remote_loc, path.as_posix())
            match (operation, stat.S_ISDIR(path_stat.st_mode or 0)):
                case (Op.CREATE, True):
                    SYNC_LOGGER.debug("Creating remote directory %s", posix_path)
                    remote.mkdir(posix_path)
                case (Op.CREATE, False) | (Op.REPLACE, False):
                    SYNC_LOGGER.debug(
                        "Uploading file %s to remote",
                        local_path / path,
                    )
                    try:
                        remote.remove(posix_path)
                    except FileNotFoundError:
                        pass
                    upload_file(
                        remote,
                        local_path / path,
                        posix_path,
                    )
                case (Op.DELETE, True):
                    # recall that for deletions, it's the *destination's* stats
                    if delete:
                        remote.rmdir(posix_path)
                case (Op.DELETE, False):
                    if delete:
                        SYNC_LOGGER.debug("Deleting remote file %s", posix_path)
                        remote.remove(posix_path)
                case op, is_dir:  # pragma: no cover
                    raise NotImplementedError(
                        f"Don't know how to handle {op} of {'directory' if is_dir else 'file'}"
                    )
